import datetime
import json
import logging
import time
import urllib.parse
import uuid
from abc import abstractmethod
from textwrap import wrap
from typing import List
from typing import Tuple

import i18n
import pymarc
from dateutil.parser import parse
from folio_uuid.folio_uuid import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient
from pymarc import Field
from pymarc import Leader
from pymarc import Record
from pymarc import Subfield

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapper_base import MapperBase
from folio_migration_tools.marc_rules_transformation.hrid_handler import HRIDHandler


class RulesMapperBase(MapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        library_configuration: LibraryConfiguration,
        task_configuration,
        schema: dict,
        conditions=None,
        parent_id_map: dict[str, tuple] = None,
    ):
        super().__init__(library_configuration, folio_client, parent_id_map)
        self.parsed_records = 0
        self.id_map: dict[str, tuple] = {}
        self.start = time.time()
        self.last_batch_time = time.time()
        self.folio_client: FolioClient = folio_client
        self.schema: dict = schema
        self.task_configuration = task_configuration
        self.conditions = conditions
        self.item_json_schema = ""
        self.mappings: dict = {}
        self.schema_properties = None
        if hasattr(self.task_configuration, "hrid_handling"):
            self.hrid_handler = HRIDHandler(
                folio_client,
                self.task_configuration.hrid_handling,
                self.migration_report,
                self.task_configuration.deactivate035_from001,
            )
        logging.info("Current user id is %s", self.folio_client.current_user)

    def print_progress(self):
        self.parsed_records += 1
        num_recs = 5000
        if self.parsed_records % num_recs == 0:
            elapsed = self.parsed_records / (time.time() - self.start)
            elapsed_formatted = "{0:.4g}".format(elapsed)
            elapsed_last = num_recs / (time.time() - self.last_batch_time)
            elapsed_formatted_last = "{0:.4g}".format(elapsed_last)
            logging.info(
                f"{elapsed_formatted_last} (avg. {elapsed_formatted}) "
                f"records/sec.\t\t{self.parsed_records:,} records processed"
            )
            self.last_batch_time = time.time()

    @abstractmethod
    def get_legacy_ids(self, marc_record: Record, idx: int):
        raise NotImplementedError()

    @staticmethod
    def dedupe_rec(rec, props_to_not_dedupe=None):
        if props_to_not_dedupe is None:
            props_to_not_dedupe = []
        # remove duplicates
        for key, value in rec.items():
            if key not in props_to_not_dedupe and isinstance(value, list):
                res = []
                for v in value:
                    if v not in res:
                        res.append(v)
                rec[key] = list(res)

    def map_field_according_to_mapping(
        self, marc_field: pymarc.Field, mappings, folio_record, legacy_ids
    ):
        for mapping in mappings:
            try:
                if "entity" not in mapping:
                    self.handle_normal_mapping(mapping, marc_field, folio_record, legacy_ids)
                else:
                    self.handle_entity_mapping(
                        marc_field,
                        mapping,
                        folio_record,
                        legacy_ids,
                    )
            except TransformationFieldMappingError as tre:
                tre.log_it()

    def handle_normal_mapping(self, mapping, marc_field: pymarc.Field, folio_record, legacy_ids):
        target = mapping["target"]
        if mapping.get("ignoreSubsequentSubfields", False):
            marc_field = self.remove_repeated_subfields(marc_field)
        if has_conditions(mapping):
            values = self.apply_rules(marc_field, mapping, legacy_ids)
            if marc_field.tag == "655":
                values[0] = f"Genre: {values[0]}"
            self.add_value_to_target(folio_record, target, values)
        elif has_value_to_add(mapping):
            value = mapping["rules"][0]["value"]
            # Stupid construct to avoid bool("false") == True
            if value == "true":
                self.add_value_to_target(folio_record, target, [True])
            elif value == "false":
                self.add_value_to_target(folio_record, target, [False])
            else:
                self.add_value_to_target(folio_record, target, [value])
        else:
            # Adding stuff without rules/Conditions.
            # Might need more complex mapping for arrays etc
            if any(mapping["subfield"]):
                values = self.handle_sub_field_delimiters(
                    ",".join(legacy_ids), mapping, marc_field
                )
                value = " ".join(values)
            else:
                value = marc_field.format_field() if marc_field else ""
            self.add_value_to_target(folio_record, target, [value])

    @staticmethod
    def set_005_as_updated_date(marc_record: Record, folio_object: dict, legacy_ids):
        try:
            f005 = marc_record["005"].data[:14]
            parsed_date = datetime.datetime.strptime(f005, "%Y%m%d%H%M%S").isoformat()
        except Exception as exception:
            if "005" in marc_record:
                Helper.log_data_issue(
                    legacy_ids,
                    f"Could not parse Last transaction date from 005 {exception}",
                    marc_record["005"].data,
                )

    @abstractmethod
    def parse_record(
        self, marc_record: Record, file_def: FileDefinition, legacy_ids: List[str]
    ) -> list[dict]:
        raise NotImplementedError()

    @staticmethod
    def use_008_for_dates(marc_record: Record, folio_object: dict, legacy_ids):
        try:
            first_six = "".join(marc_record["008"].data[:6])
            date_str = f"19{first_six}" if int(first_six[:2]) > 69 else f"20{first_six}"
            date_str_parsed = datetime.datetime.strptime(date_str, "%Y%m%d")
            if "title" in folio_object:  # only instance has titles
                folio_object["catalogedDate"] = date_str_parsed.strftime("%Y-%m-%d")
        except Exception as exception:
            if "008" in marc_record:
                Helper.log_data_issue(
                    legacy_ids,
                    f"Could not parse cat date from 008 {exception}",
                    marc_record["008"].data,
                )

    def handle_sub_field_delimiters(
        self,
        legacy_id: str,
        mapping,
        marc_field: pymarc.Field,
        condition_types: List[str] = None,
        parameter: dict = None,
    ):
        values: List[str] = []
        if mapping.get("subfield") and (custom_delimiters := mapping.get("subFieldDelimiter")):
            delimiter_map = {sub_f: " " for sub_f in mapping.get("subfield")}
            for custom_delimiter in custom_delimiters:
                delimiter_map.update(
                    {sub_f: custom_delimiter["value"] for sub_f in custom_delimiter["subfields"]}
                )
            custom_delimited_strings: List[Tuple[str, List[str]]] = []
            subfields = mapping.get("subfield")
            for custom_delimiter in custom_delimiters:
                subfields_for_delimiter = [
                    sub_f
                    for sub_f in subfields
                    if custom_delimiter["subfields"]
                    and delimiter_map[sub_f] == custom_delimiter["value"]
                ]
                subfield_collection: Tuple[str, List[str]] = (custom_delimiter["value"], [])
                subfield_collection[1].extend(marc_field.get_subfields(*subfields_for_delimiter))
                custom_delimited_strings.append(subfield_collection)
            for custom_delimited_string in custom_delimited_strings:
                if mapping.get("applyRulesOnConcatenatedData", ""):
                    values.extend(custom_delimited_string[1])
                else:
                    values.extend(
                        dict.fromkeys(
                            [
                                self.apply_rule(
                                    legacy_id,
                                    x,
                                    condition_types or [],
                                    marc_field,
                                    parameter or {},
                                )
                                for x in custom_delimited_string[1]
                            ]
                        )
                    )
                values = [custom_delimited_string[0].join(values)]
        elif mapping.get("subfield", []):
            values.extend(marc_field.get_subfields(*mapping["subfield"]))
        return values

    def get_value_from_condition(
        self,
        legacy_id,
        mapping,
        marc_field,
    ):
        stripped_conds = mapping["rules"][0]["conditions"][0]["type"].split(",")
        condition_types = list(map(str.strip, stripped_conds))
        parameter = mapping["rules"][0]["conditions"][0].get("parameter", {})
        values: List[str] = []
        if mapping.get("subfield"):
            values.extend(
                self.handle_sub_field_delimiters(
                    legacy_id, mapping, marc_field, condition_types, parameter
                )
            )
        else:
            values.append(marc_field.format_field() if marc_field else "")

        if not mapping.get("applyRulesOnConcatenatedData", "") and mapping.get("subfield", []):
            return " ".join(
                dict.fromkeys(
                    [
                        self.apply_rule(legacy_id, x, condition_types, marc_field, parameter)
                        for x in values
                    ]
                )
            )
        else:
            return self.apply_rule(
                legacy_id, " ".join(values), condition_types, marc_field, parameter
            )

    def process_marc_field(
        self,
        folio_record: dict,
        marc_field: Field,
        ignored_subsequent_fields,
        legacy_ids,
    ):
        if marc_field.tag == "880":
            mappings = self.perform_proxy_mapping(marc_field)
        else:
            tags_to_ignore = {"880", "001", "008"}
            mappings = (
                self.mappings.get(marc_field.tag, {})
                if marc_field.tag not in tags_to_ignore
                else []
            )
        if mappings:
            try:
                self.map_field_according_to_mapping(marc_field, mappings, folio_record, legacy_ids)
                if any(m.get("ignoreSubsequentFields", False) for m in mappings):
                    ignored_subsequent_fields.add(marc_field.tag)
            except Exception as ee:
                logging.error(
                    "map_field_according_to_mapping %s %s %s",
                    marc_field.tag,
                    marc_field.format_field(),
                    json.dumps(mappings),
                )
                raise ee

    def perform_proxy_mapping(self, marc_field):
        proxy_mapping = next(iter(self.mappings.get("880", [])), [])
        if "6" not in marc_field:
            self.migration_report.add("Field880Mappings", i18n.t("Records without $6"))
            return None
        if not proxy_mapping or not proxy_mapping.get("fieldReplacementBy3Digits", False):
            return None
        if not marc_field["6"][:3] or len(marc_field["6"][:3]) != 3:
            self.migration_report.add(
                "Field880Mappings", i18n.t("Records with unexpected length in $6")
            )
            return None
        first_three = marc_field["6"][:3]

        target_field = next(
            (
                r.get("targetField", "")
                for r in proxy_mapping.get("fieldReplacementRule", [])
                if r["sourceDigits"] == first_three
            ),
            first_three,
        )
        self.migration_report.add(
            "Field880Mappings",
            i18n.t("Source digits")
            + f": {marc_field['6']} "
            + i18n.t("Target field")
            + f": {target_field}",
        )
        mappings = self.mappings.get(target_field, {})
        if not mappings:
            self.migration_report.add(
                "Field880Mappings",
                i18n.t("Mapping not set up for target field")
                + f": {target_field} ({marc_field['6']})",
            )
        return mappings

    def report_marc_stats(
        self, marc_field: Field, bad_tags, legacy_ids, ignored_subsequent_fields
    ):
        self.migration_report.add("Trivia", i18n.t("Total number of Tags processed"))
        self.report_source_and_links(marc_field)
        self.report_bad_tags(marc_field, bad_tags, legacy_ids)
        mapped = marc_field.tag in self.mappings
        if marc_field.tag in ignored_subsequent_fields:
            mapped = False
        self.report_legacy_mapping(marc_field.tag, True, mapped)

    def report_source_and_links(self, marc_field: Field):
        if marc_field.is_control_field():
            return
        for subfield_2 in marc_field.get_subfields("2"):
            self.migration_report.add(
                "AuthoritySources",
                i18n.t("Source of heading or term") + f": {subfield_2.split(' ')[0]}",
            )
        for subfield_0 in marc_field.get_subfields("0"):
            code = ""
            if "(" in subfield_0 and ")" in subfield_0:
                code = subfield_0[subfield_0.find("(") + 1 : subfield_0.find(")")]
                code = code.split(" ")[0]
            elif url := urllib.parse.urlparse(subfield_0):
                if url.hostname:
                    code = subfield_0[: subfield_0.find(url.path)]
            if code:
                self.migration_report.add(
                    "AuthoritySources", i18n.t("$0 base uri or source code") + f": {code}"
                )

    def apply_rules(self, marc_field: pymarc.Field, mapping, legacy_ids):
        try:
            values = []
            value = ""
            if has_conditions(mapping):
                value = self.get_value_from_condition(",".join(legacy_ids), mapping, marc_field)
            elif has_value_to_add(mapping):
                value = mapping["rules"][0]["value"]
                if value == "false":
                    return [False]
                elif value == "true":
                    return [True]
                else:
                    return [value]
            elif not mapping.get("rules", []) or not mapping["rules"][0].get("conditions", []):
                values = self.handle_sub_field_delimiters(
                    ",".join(legacy_ids), mapping, marc_field
                )
                value = " ".join(values)
            values = wrap(value, 3) if mapping.get("subFieldSplit", "") else [value]
            return values
        except TransformationProcessError as trpe:
            self.handle_transformation_process_error(self.parsed_records, trpe)
        except TransformationFieldMappingError as fme:
            self.migration_report.add("FieldMappingErrors", fme.message)
            fme.data_value = (
                f"{fme.data_value} MARCField: {marc_field} Mapping: {json.dumps(mapping)}"
            )
            fme.log_it()
            return []
        except TransformationRecordFailedError as trfe:
            trfe.data_value = (
                f"{trfe.data_value} MARCField: {marc_field} Mapping: {json.dumps(mapping)}"
            )
            trfe.log_it()
            self.migration_report.add_general_statistics(
                i18n.t("Records failed due to an error. See data issues log for details")
            )
        except Exception as exception:
            self.handle_generic_exception(self.parsed_records, exception)

    def report_bad_tags(self, marc_field, bad_tags, legacy_ids):
        if (
            (not marc_field.tag.isnumeric())
            and marc_field.tag != "LDR"
            and marc_field.tag not in bad_tags
        ):
            self.migration_report.add("NonNumericTagsInRecord", marc_field.tag)
            message = "Non-numeric tags in records"
            Helper.log_data_issue(legacy_ids, message, marc_field.tag)
            bad_tags.add(marc_field.tag)

    def add_value_to_target(self, rec, target_string, value):
        if not value:
            return
        targets = target_string.split(".")
        if len(targets) == 1:
            self.add_value_to_first_level_target(rec, target_string, value)
        else:
            schema_parent = None
            parent = None
            schema_properties = self.schema["properties"]
            sc_prop = schema_properties
            for target in targets:  # Iterate over names in hierarcy
                if target in sc_prop:  # property is on this level
                    sc_prop = sc_prop[target]  # set current property
                else:  # next level. take the properties from the items
                    sc_prop = schema_parent["items"]["properties"][target]
                if target not in rec and not schema_parent:  # have we added this already?
                    if is_array_of_strings(sc_prop):
                        rec[target] = []
                        # break
                        # prop[target].append({})
                    elif is_array_of_objects(sc_prop):
                        rec[target] = [{}]
                        # break
                    elif (
                        schema_parent
                        and is_array_of_objects(schema_parent)
                        and sc_prop.get("type", "string") == "string"
                    ):
                        s = "This should be unreachable code. Check schema for changes"
                        logging.error(s)
                        logging.error(parent)
                        raise TransformationProcessError("", s)
                        # break
                    else:
                        if schema_parent["type"] == "array":
                            parent.append({})
                        else:
                            raise TransformationProcessError(
                                "",
                                f"Edge! Something in the schemas has changed. "
                                "The mapping of this needs to be investigated "
                                f"{target_string} {schema_properties[target_string]}",
                            )
                elif is_array_of_objects(sc_prop) and len(rec[target][-1]) == len(
                    sc_prop["items"]["properties"]
                ):
                    rec[target].append({})
                elif schema_parent and target in rec[parent][-1]:
                    rec[parent].append({})
                    if len(rec[parent][-1]) > 0:
                        rec[parent][-1][target] = value[0]
                    else:
                        rec[parent][-1] = {target: value[0]}
                elif (
                    schema_parent
                    and is_array_of_objects(schema_parent)
                    and sc_prop.get("type", "string") == "string"
                ):
                    if len(rec[parent][-1]) > 0:
                        rec[parent][-1][target] = value[0]
                    else:
                        rec[parent][-1] = {target: value[0]}
                # if target == targets[-1]:
                # prop[target] = value[0]
                # prop = rec[target]
                schema_parent = sc_prop
                parent = target

    def add_value_to_first_level_target(self, rec, target_string, value):
        sch = self.schema["properties"]
        if (
            self.task_configuration.migration_task_type == "BibsTransformer"
            and self.task_configuration.parse_cataloged_date
            and target_string == "catalogedDate"
        ):
            try:
                value = [str(parse(value[0], fuzzy=True).date())]
            except Exception as ee:
                Helper.log_data_issue("", f"Could not parse catalogedDate: {ee}", value)
                self.migration_report.add(
                    "FieldMappingErrors", i18n.t("Could not parse catalogedDate")
                )
        if not target_string or target_string not in sch:
            raise TransformationFieldMappingError(
                "",
                i18n.t("Target string '%{string}' not in Schema!", string=target_string)
                + i18n.t("Check mapping file against the schema.")
                + " "
                + i18n.t("Target type")
                + f": {sch.get(target_string,{}).get('type','')} "
                + i18n.t("Value")
                + f": {value}",
                "",
            )

        target_field = sch.get(target_string, {})
        if (
            target_field.get("type", "") == "array"
            and target_field.get("items", {}).get("type", "") == "string"
        ):
            if target_string not in rec:
                rec[target_string] = value
            else:
                rec[target_string].extend(value)

        elif target_field.get("type", "") == "string":
            if value[0]:
                rec[target_string] = value[0]
        else:
            raise TransformationProcessError(
                "",
                (
                    f"Edge! Target string: {target_string} "
                    f"Target type: {sch.get(target_string,{}).get('type','')} Value: {value}"
                ),
            )

    def remove_from_id_map(self, former_ids: List[str]):
        """removes the ID from the map in case parsing failed

        Args:
            former_ids (_type_): _description_
        """
        for former_id in [id for id in former_ids if id]:
            if former_id in self.id_map:
                del self.id_map[former_id]

    def create_entity(self, entity_mappings, marc_field: Field, entity_parent_key, index_or_legacy_id):
        entity = {}
        parent_schema_prop = self.schema.get("properties", {}).get(entity_parent_key, {})
        if parent_schema_prop.get("type", "") == "array":
            req_entity_props = parent_schema_prop.get("items", {}).get("required", [])
        elif parent_schema_prop.get("type", "") == "object":
            req_entity_props = parent_schema_prop.get("required", [])
        else:
            req_entity_props = []
        for entity_mapping in entity_mappings:
            k = entity_mapping["target"].split(".")[-1]
            if k == "authorityId" and (legacy_subfield_9 := marc_field.get("9")):
                marc_field.add_subfield("0", legacy_subfield_9)
                marc_field.delete_subfield("9")
            if my_values := [
                v
                for v in self.apply_rules(marc_field, entity_mapping, index_or_legacy_id)
                if v != ""
            ]:
                if entity_parent_key != k:
                    entity[k] = my_values[0]
                else:
                    entity = my_values[0]
            elif "alternativeMapping" in entity_mapping:
                alt_mapping = entity_mapping["alternativeMapping"]
                alt_k = alt_mapping["target"].split(".")[-1]
                if alt_values := [
                    v
                    for v in self.apply_rules(marc_field, alt_mapping, index_or_legacy_id)
                    if v != ""
                ]:
                    if entity_parent_key != alt_k:
                        entity[alt_k] = alt_values[0]
                    else:
                        entity = alt_values[0]
        missing_required_props = [
            req_entity_prop
            for req_entity_prop in req_entity_props
            if req_entity_prop not in entity
        ]
        return {} if any(missing_required_props) else entity

    def handle_entity_mapping(
        self,
        marc_field,
        mapping,
        folio_record,
        legacy_ids,
    ):
        entity_mapping = mapping["entity"]
        e_parent = entity_mapping[0]["target"].split(".")[0]
        if mapping.get("entityPerRepeatedSubfield", False):
            for temp_field in self.grouped(marc_field):
                entity = self.create_entity(entity_mapping, temp_field, e_parent, legacy_ids)
                if entity and (
                    (isinstance(entity, dict) and all(entity.values()))
                    or (isinstance(entity, list) and all(entity))
                ):
                    self.add_entity_to_record(entity, e_parent, folio_record, self.schema)
        else:
            if mapping.get("ignoreSubsequentSubfields", False):
                marc_field = self.remove_repeated_subfields(marc_field)
            entity = self.create_entity(entity_mapping, marc_field, e_parent, legacy_ids)
            if e_parent in ["precedingTitles", "succeedingTitles"]:
                self.create_preceding_succeeding_titles(entity, e_parent, folio_record["id"])
            elif entity and (
                all(
                    v
                    for k, v in entity.items()
                    if k not in ["staffOnly", "primary", "isbnValue", "issnValue"]
                )
                or e_parent in ["electronicAccess", "publication"]
                or (
                    e_parent.startswith("holdingsStatements") and any(v for k, v in entity.items())
                )
            ):
                self.add_entity_to_record(entity, e_parent, folio_record, self.schema)
            else:
                sfs = " - ".join(
                    f"{f[0]}:{('has_value' if f[1].strip() else 'empty')}" for f in marc_field
                )
                pattern = " - ".join(f"{k}:'{bool(v)}'" for k, v in entity.items())
                self.migration_report.add(
                    "IncompleteEntityMapping",
                    f"{marc_field.tag} {sfs} ->>-->> {e_parent} {pattern}  ",
                )
                # Experimental
                # self.add_entity_to_record(entity, e_parent, rec, self.schema)

    def handle_suppression(
        self, folio_record, file_def: FileDefinition, only_discovery_suppress: bool = False
    ):
        folio_record["discoverySuppress"] = file_def.discovery_suppressed
        self.migration_report.add(
            "Suppression",
            i18n.t("Suppressed from discovery") + f' = {folio_record["discoverySuppress"]}',
        )
        if not only_discovery_suppress:
            folio_record["staffSuppress"] = file_def.staff_suppressed
            self.migration_report.add(
                "Suppression", i18n.t("Staff suppressed") + f' = {folio_record["staffSuppress"]} '
            )

    def create_preceding_succeeding_titles(self, entity, e_parent, identifier):
        self.migration_report.add("PrecedingSuccedingTitles", f"{e_parent} " + i18n.t("created"))
        # TODO: Make these uuids deterministic
        new_entity = {
            "id": str(uuid.uuid4()),
            "title": entity.get("title"),
            "identifiers": [],
        }
        if e_parent == "precedingTitles":
            new_entity["succeedingInstanceId"] = identifier
        else:
            new_entity["precedingInstanceId"] = identifier
        if new_entity.get("isbnValue", ""):
            new_entity["identifiers"].append(
                {
                    "identifierTypeId": new_entity.get("isbnId"),
                    "value": new_entity.get("isbnValue"),
                }
            )
        if new_entity.get("issnValue", ""):
            new_entity["identifiers"].append(
                {
                    "identifierTypeId": new_entity.get("issnId"),
                    "value": new_entity.get("issnValue"),
                }
            )
        self.extradata_writer.write(e_parent, new_entity)

    def apply_rule(self, legacy_id, value, condition_types, marc_field, parameter):
        v = value
        for condition_type in iter(condition_types):
            try:
                v = self.conditions.get_condition(
                    condition_type, legacy_id, v, parameter, marc_field
                )
            except AttributeError as attr_error:
                raise TransformationProcessError(
                    legacy_id, attr_error, condition_type
                ) from attr_error
        return v

    @staticmethod
    def add_entity_to_record(entity, entity_parent_key, rec, schema):
        sch = schema["properties"]
        if sch[entity_parent_key]["type"] == "array":
            if entity_parent_key not in rec:
                rec[entity_parent_key] = [entity]
            else:
                rec[entity_parent_key].append(entity)
        else:
            rec[entity_parent_key] = entity

    @staticmethod
    def grouped(marc_field: Field):
        """Groups the subfields
        s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,s2n+1,s2n+2,...s3n-1), ...


        Args:
            marc_field (Field): _description_

        Returns:
            _type_: _description_
        """
        unique_subfields: list = []
        repeated_subfields: list = []
        results = []
        for sf, sf_vals in marc_field.subfields_as_dict().items():
            if len(sf_vals) == 1:
                unique_subfields.append(Subfield(code=sf, value=sf_vals[0]))
            else:
                repeated_subfields.extend([Subfield(code=sf, value=sf_val) for sf_val in sf_vals])
        if any(repeated_subfields):
            for repeated_subfield in repeated_subfields:
                new_subfields = [repeated_subfield]
                new_subfields.extend(unique_subfields)
                temp_field = Field(
                    tag=marc_field.tag,
                    indicators=marc_field.indicators,
                    subfields=new_subfields,
                )
                results.append(temp_field)
        else:
            temp_field = Field(
                tag=marc_field.tag,
                indicators=marc_field.indicators,
                subfields=unique_subfields,
            )
            results.append(temp_field)
        return results

    @staticmethod
    def remove_repeated_subfields(marc_field: Field):
        """Removes repeated subfields
        s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,s2n+1,s2n+2,...s3n-1), ...

        Args:
            marc_field (Field): _description_

        Returns:
            _type_: _description_
        """
        new_subfields = []
        for sf, sf_vals in marc_field.subfields_as_dict().items():
            new_subfields.extend([Subfield(code=sf, value=sf_vals[0])])
        return Field(
            tag=marc_field.tag,
            indicators=marc_field.indicators,
            subfields=new_subfields,
        )

    @staticmethod
    def save_source_record(
        srs_records_file,
        record_type: FOLIONamespaces,
        folio_client: FolioClient,
        marc_record: Record,
        folio_record,
        legacy_ids: List[str],
        suppress: bool,
    ):
        """Saves the source Marc_record to the Source record Storage module

        Args:
            srs_records_file (_type_): _description_
            record_type (FOLIONamespaces): _description_
            folio_client (FolioClient): _description_
            marc_record (Record): _description_
            folio_record (_type_): _description_
            legacy_ids (List[str]): _description_
            suppress (bool): _description_
        """
        srs_id = RulesMapperBase.create_srs_id(record_type, folio_client.okapi_url, legacy_ids[-1])

        marc_record.add_ordered_field(
            Field(
                tag="999",
                indicators=["f", "f"],
                subfields=[
                    Subfield(code="i", value=folio_record["id"]),
                    Subfield(code="s", value=srs_id),
                ],
            )
        )
        # Since they all should be UTF encoded, make the leader align.
        try:
            temp_leader = Leader(marc_record.leader)
            temp_leader[9] = "a"
            marc_record.leader = temp_leader
        except Exception as ee:
            logging.exception(
                "Something is wrong with the marc records leader: %s, %s", marc_record.leader, ee
            )
        srs_record_string = RulesMapperBase.get_srs_string(
            marc_record,
            folio_record,
            srs_id,
            suppress,
            record_type,
        )
        srs_records_file.write(f"{srs_record_string}\n")

    @staticmethod
    def create_srs_id(record_type, okapi_url: str, legacy_id: str):
        srs_types = {
            FOLIONamespaces.holdings: FOLIONamespaces.srs_records_holdingsrecord,
            FOLIONamespaces.instances: FOLIONamespaces.srs_records_bib,
            FOLIONamespaces.authorities: FOLIONamespaces.srs_records_auth,
            FOLIONamespaces.edifact: FOLIONamespaces.srs_records_edifact,
        }

        return str(FolioUUID(okapi_url, srs_types.get(record_type), legacy_id))

    @staticmethod
    def get_bib_id_from_907y(marc_record: Record, index_or_legacy_id):
        try:
            return list(set(marc_record["907"].get_subfields("a", "y")))
        except Exception as e:
            raise TransformationRecordFailedError(
                index_or_legacy_id,
                (
                    "907 $y and $a is missing is missing, although they is "
                    "required for this legacy ILS choice"
                ),
                marc_record.as_json(),
            ) from e

    @staticmethod
    def get_bib_id_from_990a(marc_record: Record, index_or_legacy_id):
        res = {f["a"].strip() for f in marc_record.get_fields("990") if "a" in f}
        if marc_record["001"].format_field().strip():
            res.add(marc_record["001"].format_field().strip())
        if any(res):
            return list(res)
        else:
            raise TransformationRecordFailedError(
                index_or_legacy_id,
                "neither 990$a or 001 found in record.",
                marc_record.as_json(),
            )

    @staticmethod
    def get_bib_id_from_001(marc_record: Record, index_or_legacy_id):
        try:
            return [marc_record["001"].format_field().strip()]
        except Exception as e:
            raise TransformationRecordFailedError(
                index_or_legacy_id,
                "001 is missing, although it is required for Voyager migrations",
                marc_record.as_json(),
            ) from e

    @staticmethod
    def get_srs_string(
        marc_record: Record,
        folio_object: dict,
        srs_id,
        discovery_suppress: bool,
        record_type: FOLIONamespaces,
    ):
        record_types = {
            FOLIONamespaces.holdings: "MARC_HOLDING",
            FOLIONamespaces.instances: "MARC_BIB",
            FOLIONamespaces.authorities: "MARC_AUTHORITY",
            FOLIONamespaces.edifact: "EDIFACT",
        }

        id_holders = {
            FOLIONamespaces.instances: {
                "instanceId": folio_object["id"],
                "instanceHrid": folio_object.get("hrid", ""),
            },
            FOLIONamespaces.holdings: {
                "holdingsId": folio_object["id"],
                "holdingsHrid": folio_object.get("hrid", ""),
            },
            FOLIONamespaces.authorities: {
                "authorityId": folio_object["id"],
                "authorityHrid": marc_record["001"].data,
            },
            FOLIONamespaces.edifact: {},
        }

        my_tuple_json = marc_record.as_json()
        raw_record = {"id": srs_id, "content": my_tuple_json}
        parsed_record = {"id": srs_id, "content": json.loads(my_tuple_json)}
        record = {
            "id": srs_id,
            "deleted": False,
            "matchedId": srs_id,
            "generation": 0,
            "recordType": record_types.get(record_type),
            "rawRecord": raw_record,
            "parsedRecord": parsed_record,
            "additionalInfo": {"suppressDiscovery": discovery_suppress},
            "externalIdsHolder": id_holders.get(record_type),
            "state": "ACTUAL",
            "leaderRecordStatus": parsed_record["content"]["leader"][5]
            if parsed_record["content"]["leader"][5] in [*"acdnposx"]
            else "d",
        }
        return json.dumps(record)


def has_conditions(mapping):
    return mapping.get("rules", []) and mapping["rules"][0].get("conditions", [])


def has_value_to_add(mapping):
    return mapping.get("rules", []) and mapping["rules"][0].get("value", "")


def is_array_of_strings(schema_property):
    sc_prop_type = schema_property.get("type", "string")
    return sc_prop_type == "array" and schema_property["items"]["type"] == "string"


def is_array_of_objects(schema_property):
    sc_prop_type = schema_property.get("type", "string")
    return sc_prop_type == "array" and schema_property["items"]["type"] == "object"



import json
import logging
import os
from pathlib import Path
from typing import List

from folio_migration_tools.custom_exceptions import TransformationProcessError


class ExtradataWriter:
    __instance = None
    __inited = False

    def __new__(cls, path_to_file: Path) -> "ExtradataWriter":
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __init__(self, path_to_file: Path) -> None:
        if type(self).__inited:
            return
        self.cache: List[str] = []
        self.path_to_file: Path = path_to_file
        if self.path_to_file.is_file():
            os.remove(self.path_to_file)
        type(self).__inited = True

    def write(self, record_type: str, data_to_write: dict, flush=False):
        try:
            if data_to_write:
                self.cache.append(f"{record_type}\t{json.dumps(data_to_write)}\n")
            if len(self.cache) > 1000 or flush:
                with open(self.path_to_file, "a") as extradata_file:
                    extradata_file.writelines(self.cache)
                    self.cache = []
                    logging.debug("Extradata writer flushing the cache")
        except Exception as ee:
            error_message = "Something went wrong in extradata Writer"
            logging.error(error_message)
            raise TransformationProcessError("", error_message, record_type) from ee

    def flush(self):
        self.write("", {}, True)
        if self.path_to_file.is_file() and os.stat(self.path_to_file).st_size == 0:
            logging.info("Removing extradata file since it is empty")
            os.remove(self.path_to_file)



import logging
import re

import i18n
import pymarc
from folioclient import FolioClient
from pymarc import field

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FolioRelease
from folio_migration_tools.marc_rules_transformation.rules_mapper_base import (
    RulesMapperBase,
)

# flake8: noqa: s


class Conditions:
    holdings_type_map = {
        "u": "Unknown",
        "v": "Multi-part monograph",
        "x": "Monograph",
        "y": "Serial",
    }

    def __init__(
        self,
        folio: FolioClient,
        mapper: RulesMapperBase,
        object_type,
        folio_release: FolioRelease,
        default_call_number_type_name="",
    ):
        self.filter_chars = r"[.,\/#!$%\^&\*;:{}=\-_`~()]"
        self.filter_chars_dop = r"[.,\/#!$%\^&\*;:{}=\_`~()]"
        self.folio_release: FolioRelease = folio_release
        self.filter_last_chars = r",$"
        self.folio = folio
        self.default_contributor_type = ""
        self.mapper = mapper
        self.ref_data_dicts = {}
        if object_type == "bibs":
            self.setup_reference_data_for_all()
            self.setup_reference_data_for_bibs()
        elif object_type == "auth":
            self.setup_reference_data_for_auth()
        else:
            self.setup_reference_data_for_all()
            self.setup_reference_data_for_items_and_holdings(default_call_number_type_name)
        self.condition_cache: dict = {}

    def setup_reference_data_for_bibs(self):
        logging.info("Setting up reference data for bib transformation")
        logging.info("%s\tcontrib_name_types", len(self.folio.contrib_name_types))
        logging.info("%s\tcontributor_types", len(self.folio.contributor_types))
        logging.info("%s\talt_title_types", len(self.folio.alt_title_types))
        logging.info("%s\tidentifier_types", len(self.folio.identifier_types))
        # Raise for empty settings
        if not self.folio.contributor_types:
            raise TransformationProcessError("", "No contributor_types in FOLIO")
        if not self.folio.contrib_name_types:
            raise TransformationProcessError("", "No contributor name types in FOLIO")
        if not self.folio.identifier_types:
            raise TransformationProcessError("", "No identifier_types in FOLIO")
        if not self.folio.alt_title_types:
            raise TransformationProcessError("", "No alt_title_types in FOLIO")

        # Set defaults
        logging.info("Setting defaults")
        self.default_contributor_name_type = self.folio.contrib_name_types[0]["id"]
        logging.info("Contributor name type:\t%s", self.default_contributor_name_type)
        self.default_contributor_type = next(
            ct for ct in self.folio.contributor_types if ct["code"] == "ctb"
        )
        logging.info("Contributor type:\t%s", self.default_contributor_type["id"])

    def setup_reference_data_for_items_and_holdings(self, default_call_number_type_name):
        logging.info(f"{len(self.folio.locations)}\tlocations")
        self.default_call_number_type = {}
        logging.info("%s\tholding_note_types", len(self.folio.holding_note_types))
        logging.info("%s\tcall_number_types", len(self.folio.call_number_types))
        self.setup_and_validate_holdings_types()
        # Raise for empty settings
        if not self.folio.holding_note_types:
            raise TransformationProcessError("", "No holding_note_types in FOLIO")
        if not self.folio.call_number_types:
            raise TransformationProcessError("", "No call_number_types in FOLIO")
        if not self.folio.locations:
            raise TransformationProcessError("", "No locations in FOLIO")

        # Set defaults
        logging.info("Defaults")
        self.default_call_number_type = next(
            (
                ct
                for ct in self.folio.call_number_types
                if ct["name"] == default_call_number_type_name
            ),
            None,
        )
        if not self.default_call_number_type:
            raise TransformationProcessError(
                "",
                (
                    f"No callnumber type with name "
                    f"{default_call_number_type_name} in FOLIO.\n"
                    "Please specify another UUID as the default Callnumber Type"
                ),
            )
        logging.info("Default Callnumber type Name:\t%s", self.default_call_number_type["name"])

    def setup_and_validate_holdings_types(self):
        self.holdings_types = list(
            self.folio.folio_get_all("/holdings-types", "holdingsTypes", "", 1000)
        )
        if not self.holdings_types:
            raise TransformationProcessError("", "No holdings_types in FOLIO")
        missing_holdings_types = [
            ht
            for ht in self.holdings_type_map.values()
            if ht not in [ht_ref["name"] for ht_ref in self.holdings_types]
        ]
        if any(missing_holdings_types):
            raise TransformationProcessError(
                "",
                "Holdings types are missing from the tenant. Please set them up",
                missing_holdings_types,
            )
        logging.info("%s\tholdings types", len(self.holdings_types))

    def setup_reference_data_for_all(self):
        logging.info(f"{len(self.folio.class_types)}\tclass_types")
        logging.info(
            f"{len(self.folio.electronic_access_relationships)}\telectronic_access_relationships"
        )
        self.statistical_codes = self.folio.statistical_codes
        logging.info(f"{len(self.statistical_codes)} \tstatistical_codes")

        # Raise for empty settings
        if not self.folio.class_types:
            raise TransformationProcessError("", "No class_types in FOLIO")

    def setup_reference_data_for_auth(self):
        self.authority_note_types = list(
            self.folio.folio_get_all(
                "/authority-note-types", "authorityNoteTypes", self.folio.cql_all, 1000
            )
        )
        logging.info(f"{len(self.authority_note_types)} \tAuthority note types")
        logging.info(f"{len(self.folio.identifier_types)} \tidentifier types")

    def get_condition(
        self, name, legacy_id, value, parameter=None, marc_field: field.Field = None
    ):
        try:
            return self.condition_cache.get(name)(legacy_id, value, parameter, marc_field)
        # Exception should only handle the missing condition from the cache.
        # All other exceptions should propagate up
        except Exception:
            attr = getattr(self, "condition_" + str(name))
            self.condition_cache[name] = attr
            return attr(legacy_id, value, parameter, marc_field)

    def condition_trim_punctuation(self, legacy_id, value, parameter, marc_field: field.Field):
        """
        Strip leading and trailing whitespace, as well as any trailing commas or periods, unless
        the period is preceded by a single alpha character (eg. "John D."). Also preserves any
        trailing "-" (eg. "1981-"). This condition was introduced in Poppy.
        """
        pattern1 = re.compile(r"^(.*?)\s.[.]$")
        pattern2 = re.compile(r"^(.*?)\s.,[.]$")
        value = value.strip()
        if pattern1.match(value) or value.endswith("-"):
            return value
        elif pattern2.match(value):
            return value.rstrip(",")
        elif value.endswith(".") or value.endswith(","):
            return value[:-1]
        return value

    def condition_trim_period(self, legacy_id, value, parameter, marc_field: field.Field):
        return value.strip().rstrip(".").rstrip(",")

    def condition_trim(self, legacy_id, value, parameter, marc_field: field.Field):
        return value.strip()

    def condition_set_contributor_type_id_by_code_or_name(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        contributor_code_subfield = parameter.get("contributorCodeSubfield", "4")
        for subfield in marc_field.get_subfields(contributor_code_subfield):
            normalized_subfield = re.sub(r"[^A-Za-z0-9 ]+", "", subfield.strip())
            t = self.get_ref_data_tuple_by_code(
                self.folio.contributor_types, "contrib_types_c", normalized_subfield
            )
            if not t:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    (
                        f'Mapping failed for ${contributor_code_subfield} "{subfield}" '
                        f"({normalized_subfield}) "
                    ),
                )
                Helper.log_data_issue(
                    legacy_id,
                    f"Mapping failed for ${contributor_code_subfield}",
                    f'{subfield}" ({normalized_subfield}) ',
                )
            else:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    (
                        i18n.t(
                            'Contributor type code "%{code}" found for $%{code_subfield}',
                            code=t[1],
                            code_subfield=contributor_code_subfield,
                        )
                        + f' "{subfield}" ({normalized_subfield}))'
                    ),
                )
                return t[0]
        fallback_name_field = "j" if marc_field.tag in ["111", "711"] else "e"
        contributor_name_subfield = parameter.get("contributorNameSubfield", fallback_name_field)
        for subfield in marc_field.get_subfields(contributor_name_subfield):
            normalized_subfield = re.sub(r"[^A-Za-z0-9 ]+", "", subfield.strip())
            t = self.get_ref_data_tuple_by_name(
                self.folio.contributor_types, "contrib_types_n", normalized_subfield
            )

            if not t:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    (
                        f"Mapping failed for {marc_field.tag} ${contributor_name_subfield} "
                        f"{subfield} (Normalized: {normalized_subfield}) "
                    ),
                )
                Helper.log_data_issue(
                    legacy_id,
                    f"Mapping failed for {marc_field.tag} ${contributor_name_subfield}",
                    f'{subfield}" ({normalized_subfield}) ',
                )
            else:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    (
                        f"Contributor type name {t[1]} found for {marc_field.tag} "
                        f"${contributor_name_subfield} {normalized_subfield} ({subfield}) "
                    ),
                )
                return t[0]
        return ""

    def condition_set_holdings_type_id(self, legacy_id, value, parameter, marc_field: field.Field):
        self.mapper.migration_report.add("HoldingsTypeMapping", i18n.t("Condition in rules hit"))
        return ""

    def condition_concat_subfields_by_name(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        subfields_to_concat = parameter.get("subfieldsToConcat", [])
        subfields_to_stop_concat = parameter.get("subfieldsToStopConcat", [])
        concat_subfields = []
        subfields = marc_field.subfields
        for t in subfields:
            if t[0] in subfields_to_stop_concat:
                break
            elif t[0] in subfields_to_concat:
                concat_subfields.append(t[1])
        concat_string = " ".join(concat_subfields)
        return f"{value} {concat_string}"

    def condition_get_value_if_subfield_is_empty(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        if value.strip():
            return value.strip()
        self.mapper.migration_report.add(
            "AddedValueFromParameter",
            i18n.t(
                "Tag: %{tag}. Added value: %{value}", tag=marc_field.tag, value=parameter["value"]
            ),
        )
        return parameter["value"]

    def condition_remove_ending_punc(self, legacy_id, value, parameter, marc_field: field.Field):
        v = value
        chars = ".;:,/+=- "
        while any(v) > 0 and v[-1] in chars:
            v = v.rstrip(v[-1])
        return v

    def condition_set_instance_format_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        # This method only handles the simple case of 2-character codes of RDA in the first 338$b
        # Other cases are handled in performAddidtionalParsing in the mapper class
        try:
            t = self.get_ref_data_tuple_by_code(
                self.folio.instance_formats, "instance_formats_code", value
            )
            self.mapper.migration_report.add(
                "InstanceFormat",
                i18n.t("Successful match") + f'  - "{value}"->{t[1]}',
            )
            return t[0]
        except Exception:
            self.mapper.migration_report.add(
                "InstanceFormat",
                i18n.t('Code from 338$b NOT found in FOLIO: "%{value}"', value=value),
            )

            return ""

    def condition_remove_prefix_by_indicator(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        """Returns the index title according to the rules"""
        ind2 = marc_field.indicator2
        reg_str = r"[\s:\/]{0,3}$"
        if ind2 not in map(str, range(1, 9)):
            return re.sub(reg_str, "", value)

        num_take = int(ind2)
        return re.sub(reg_str, "", value[num_take:])

    def condition_capitalize(self, legacy_id, value, parameter, marc_field: field.Field):
        return value.capitalize()

    def condition_clean_isbn(self, legacy_id, value, parameter, marc_field: field.Field):
        return value

    def condition_set_issuance_mode_id(self, legacy_id, value, parameter, marc_field: field.Field):
        # mode of issuance is handled elsewhere in the mapping.
        return ""

    def condition_set_publisher_role(self, legacy_id, value, parameter, marc_field: field.Field):
        roles = {
            "0": "Production",
            "1": "Publication",
            "2": "Distribution",
            "3": "Manufacture",
            "4": "Copyright notice date",
        }
        role = roles.get(marc_field.indicator2, "")
        self.mapper.migration_report.add(
            "MappedPublisherRoleFromIndicator2",
            f"{marc_field.tag} ind2 {marc_field.indicator2}->{role}",
        )
        return role

    def condition_set_identifier_type_id_by_value(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        if "oclc_regex" in parameter:
            if re.match(parameter["oclc_regex"], value):
                t = self.get_ref_data_tuple_by_name(
                    self.folio.identifier_types,
                    "identifier_types",
                    parameter["names"][1],
                )
            else:
                t = self.get_ref_data_tuple_by_name(
                    self.folio.identifier_types,
                    "identifier_types",
                    parameter["names"][0],
                )
            self.mapper.migration_report.add(
                "MappedIdentifierTypes", f"{marc_field.tag} -> {t[1]}"
            )
            return t[0]
        identifier_type = next(
            (
                f
                for f in self.folio.identifier_types
                if (
                    f["name"] in parameter.get("names", "non existant")
                    or f["name"] in parameter.get("name", "non existant")
                )
            ),
            None,
        )
        self.mapper.migration_report.add("MappedIdentifierTypes", identifier_type["name"])
        my_id = identifier_type["id"]
        if not my_id:
            raise TransformationFieldMappingError(
                legacy_id,
                i18n.t("no matching identifier_types in %{names}", names=parameter["names"]),
                marc_field,
            )
        return my_id

    def condition_set_holding_note_type_id_by_name(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        self.mapper.migration_report.add(
            "Exceptions",
            (
                "Condition set_holding_note_type_id_by_name is deprecated. "
                "Use set_holdings_note_type_id instead"
            ),
        )
        return self.condition_set_holdings_note_type_id(legacy_id, value, parameter, marc_field)

    def condition_set_holdings_note_type_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.folio.holding_note_types, "holding_note_types", parameter["name"]
            )
            self.mapper.migration_report.add("MappedNoteTypes", t[1])
            return t[0]
        except Exception as ee:
            logging.error(ee)
            raise TransformationRecordFailedError(
                legacy_id,
                f'Holdings note type mapping error.\tParameter: {parameter.get("name", "")}\t'
                f"MARC Field: {marc_field}. Is mapping rules and ref data aligned?",
                parameter.get("name", ""),
            ) from ee

    def condition_set_authority_note_type_id(
        self, legacy_id, _, parameter, marc_field: field.Field
    ):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.authority_note_types, "authority_note_types", parameter["name"]
            )
            self.mapper.migration_report.add("MappedNoteTypes", t[1])
            return t[0]
        except Exception as ee:
            logging.error(ee)
            raise TransformationProcessError(
                legacy_id,
                f'Authority note type mapping error.\tParameter: {parameter.get("name", "")}\t'
                f"MARC Field: {marc_field}. Is mapping rules and ref data aligned?",
                parameter.get("name", ""),
            ) from ee

    def condition_set_classification_type_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.folio.class_types, "class_types", parameter["name"]
            )
            self.mapper.migration_report.add("MappedClassificationTypes", t[1])
            return t[0]
        except Exception:
            raise TransformationRecordFailedError(
                legacy_id,
                f'Classification mapping error.\tParameter: "{parameter.get("name", "")}"\t'
                f"MARC Field: {marc_field}. Is mapping rules and ref data aligned?",
                parameter.get("name", ""),
            )

    def condition_char_select(self, legacy_id, value, parameter, marc_field: field.Field):
        return value[parameter["from"] : parameter["to"]]

    def condition_set_receipt_status(self, legacy_id, value, parameter, marc_field: field.Field):
        if len(value) < 7:
            self.mapper.migration_report.add(
                "ReceiptStatusMapping", i18n.t("008 is too short") + f": {value}"
            )
            return ""
        try:
            status_map = {
                "0": "Unknown",
                "1": "Other receipt or acquisition status",
                "2": "Received and complete or ceased",
                "3": "On order",
                "4": "Currently received",
                "5": "Not currently received",
                "6": "External access",
            }
            mapped_value = status_map[value[6]]
            self.mapper.migration_report.add(
                "ReceiptStatusMapping",
                i18n.t(
                    "%{value} mapped to %{mapped_value}", value=value[6], mapped_value=mapped_value
                ),
            )

            return
        except Exception:
            self.mapper.migration_report.add(
                "ReceiptStatusMapping", i18n.t("%{value} not found in map.", value=value)
            )
            return "Unknown"

    def condition_set_identifier_type_id_by_name(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.folio.identifier_types, "identifier_types", parameter["name"]
            )
            self.mapper.migration_report.add(
                "MappedIdentifierTypes", f"{marc_field.tag} -> {t[1]}"
            )
            return t[0]
        except Exception as ee:
            logging.exception("Identifier type")
            raise TransformationProcessError(
                legacy_id,
                f'Unmapped identifier type : "{parameter["name"]}"\tMARC Field: {marc_field}'
                f"MARC Field: {marc_field}. Is mapping rules and ref data aligned? ",
                {parameter["name"]},
            ) from ee

    def condition_set_contributor_name_type_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.folio.contrib_name_types, "contrib_name_types", parameter["name"]
            )
            self.mapper.migration_report.add(
                "MappedContributorNameTypes", f"{marc_field.tag} -> {t[1]}"
            )
            return t[0]
        except Exception:
            self.mapper.migration_report.add("UnmappedContributorNameTypes", parameter["name"])
            return self.default_contributor_name_type

    def condition_set_note_type_id(self, legacy_id, value, parameter, marc_field: field.Field):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.folio.instance_note_types, "instance_not_types", parameter["name"]
            )
            self.mapper.migration_report.add(
                "MappedNoteTypes",
                f"{marc_field.tag} ({parameter.get('name', '')}) -> {t[1]}",
            )
            return t[0]
        except Exception:
            raise ValueError(f"Instance note type not found for {marc_field} {parameter}")

    def condition_set_contributor_type_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        for subfield in marc_field.get_subfields("4"):
            normalized_subfield = re.sub(r"[^A-Za-z0-9 ]+", "", subfield.strip())
            t = self.get_ref_data_tuple_by_code(
                self.folio.contributor_types, "contrib_types_c", normalized_subfield
            )
            if not t:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    i18n.t(
                        'Mapping failed for %{tag} "%{subfield}" (%{normalized_subfield})',
                        tag="$4",
                        subfield=subfield,
                        normalized_subfield=normalized_subfield,
                    ),
                )
                Helper.log_data_issue(
                    legacy_id,
                    "Mapping failed for $4",
                    f'{subfield}" ({normalized_subfield}) ',
                )
            else:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    i18n.t(
                        'Contributor type code "%{code}" found for $%{code_subfield}',
                        code=t[1],
                        code_subfield="4",
                        normalized_subfield=normalized_subfield,
                    )
                    + f' "%{subfield}" (%{normalized_subfield}))',
                )
                return t[0]
        subfield_code = "j" if marc_field.tag in ["111", "711"] else "e"
        for subfield in marc_field.get_subfields(subfield_code):
            normalized_subfield = re.sub(r"[^A-Za-z0-9 ]+", "", subfield.strip())
            t = self.get_ref_data_tuple_by_name(
                self.folio.contributor_types, "contrib_types_n", normalized_subfield
            )

            if not t:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    i18n.t(
                        'Mapping failed for %{tag} "%{subfield}" (Normalized: %{normalized_subfield})',
                        tag=f"{marc_field.tag} $e",
                        subfield=subfield,
                        normalized_subfield=normalized_subfield,
                    ),
                )
                Helper.log_data_issue(
                    legacy_id,
                    f"Mapping failed for {marc_field.tag} $e",
                    f'{subfield}" ({normalized_subfield}) ',
                )
            else:
                self.mapper.migration_report.add(
                    "ContributorTypeMapping",
                    i18n.t(
                        "Contributor type name %{name} found for %{tag}",
                        name=t[1],
                        tag=marc_field.tag,
                    )
                    + f" $e {normalized_subfield} ({subfield}) ",
                )
                return t[0]
        return self.default_contributor_type["id"]

    def condition_set_url_relationship(self, legacy_id, value, parameter, marc_field: field.Field):
        return self._extracted_from_condition_set_electronic_access_relations_id_2("8", marc_field)

    def condition_set_call_number_type_by_indicator(
        self, legacy_id, value, parameter, marc_field: pymarc.Field
    ):
        self.mapper.migration_report.add(
            "Exceptions",
            (
                "Condition set_call_number_type_by_indicator is deprecated. "
                "Change to set_call_number_type_id"
            ),
        )
        return self.condition_set_call_number_type_id(legacy_id, value, parameter, marc_field)

    def condition_set_call_number_type_id(
        self, legacy_id, value, parameter, marc_field: pymarc.Field
    ):
        first_level_map = {
            "0": "Library of Congress classification",
            "1": "Dewey Decimal classification",
            "2": "National Library of Medicine classification",
            "3": "Superintendent of Documents classification",
            "4": "Shelving control number",
            "5": "Title",
            "6": "Shelved separately",
            "7": "Source specified in subfield $2",
            "8": "Other scheme",
        }

        # CallNumber type specified in $2. This needs further mapping
        if marc_field.indicator1 == "7" and "2" in marc_field:
            self.mapper.migration_report.add(
                "CallNumberTypeMapping",
                i18n.t("Unhandled call number type in $2 (ind1 == 7)" + str(marc_field["2"])),
            )
            return self.default_call_number_type["id"]

        # Normal way. Type in ind1
        call_number_type_name_temp = first_level_map.get(marc_field.indicator1, "")
        if not call_number_type_name_temp:
            self.mapper.migration_report.add(
                "CallNumberTypeMapping",
                (
                    i18n.t(
                        'Unhandled call number type in ind1: "%{ind1}".\n Returning default Callnumber type: %{type}',
                        ind1=marc_field.indicator1,
                        type=self.default_call_number_type["name"],
                    )
                ),
            )
            return self.default_call_number_type["id"]
        t = self.get_ref_data_tuple_by_name(
            self.folio.call_number_types, "cnt", call_number_type_name_temp
        )
        if t:
            self.mapper.migration_report.add(
                "CallNumberTypeMapping",
                i18n.t("Mapped from Indicator 1") + f" {marc_field.indicator1} -> {t[1]}",
            )
            return t[0]

        self.mapper.migration_report.add(
            "CallNumberTypeMapping",
            (
                "Mapping failed. Setting default CallNumber type: "
                f'{self.default_call_number_type["name"]}'
            ),
        )

        return self.default_call_number_type["id"]

    def condition_set_contributor_type_text(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        for subfield in marc_field.get_subfields("4", "e"):
            normalized_subfield = re.sub(r"[^A-Za-z0-9 ]+", "", subfield.strip())
            for cont_type in self.folio.contributor_types:
                if normalized_subfield in [cont_type["code"], cont_type["name"]]:
                    return cont_type["name"]
        try:
            return value
        except IndexError as ee:
            return ""

    def condition_set_alternative_title_type_id(self, legacy_id, value, parameter, marc_field):
        try:
            t = self.get_ref_data_tuple_by_name(
                self.folio.alt_title_types, "alt_title_types", parameter["name"]
            )
            self.mapper.migration_report.add("MappedAlternativeTitleTypes", t[1])
            return t[0]
        except Exception:
            raise TransformationProcessError(
                legacy_id,
                f"Alternative title type not found for {parameter['name']} {marc_field}",
            )

    def condition_set_location_id_by_code(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        self.mapper.migration_report.add(
            "Exceptions",
            (
                "set_location_id_by_code condition used in rules. "
                "Deprecated condition. Switch to set_permanent_location_id"
            ),
        )
        return self.condition_set_permanent_location_id(legacy_id, value, parameter, marc_field)

    def condition_set_permanent_location_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        if "legacy_locations" not in self.ref_data_dicts:
            try:
                d = {lm["legacy_code"]: lm["folio_code"] for lm in self.mapper.location_map}
                self.ref_data_dicts["legacy_locations"] = d
                for folio_code in d.values():
                    t = self.get_ref_data_tuple_by_code(
                        self.folio.locations, "locations", folio_code
                    )
                    if not t:
                        raise TransformationProcessError(
                            "", "No FOLIO location found for code", folio_code
                        )
                if "*" not in d:
                    raise TransformationProcessError(
                        "",
                        (
                            "Fallback location mapping missing. Add a row with a * in the "
                            "legacy_code column and a location code to map unmapped locations to"
                        ),
                        "",
                    )
            except KeyError as ke:
                if "folio_code" in str(ke):
                    raise TransformationProcessError(
                        legacy_id, "Your location map lacks the column folio_code"
                    ) from ke
                if "legacy_code" in str(ke):
                    raise TransformationProcessError(
                        legacy_id, "Your location map lacks the column legacy_code"
                    ) from ke

        # Get the right code from the location map
        mapped_code = self.ref_data_dicts["legacy_locations"].get(value.strip(), "").strip()
        if not mapped_code:
            mapped_code = self.ref_data_dicts["legacy_locations"].get("*", "").strip()
            if mapped_code:
                self.mapper.migration_report.add(
                    "LocationMapping", i18n.t("Fallback mapping") + f": {value}->{mapped_code}"
                )
        # Get the FOLIO UUID for the code and return it
        t = self.get_ref_data_tuple_by_code(self.folio.locations, "locations", mapped_code)
        if not t:
            self.mapper.migration_report.add(
                "LocationMapping", i18n.t("Unmapped code") + f": '{value}'"
            )
            raise TransformationRecordFailedError(
                legacy_id, "Could not map location from legacy code", value
            )
        self.mapper.migration_report.add("LocationMapping", f"'{value}' ({mapped_code}) -> {t[1]}")
        return t[0]

    def get_ref_data_tuple_by_code(self, ref_data, ref_name, code):
        return self.get_ref_data_tuple(ref_data, ref_name, code, "code")

    def get_ref_data_tuple_by_name(self, ref_data, ref_name, name):
        return self.get_ref_data_tuple(ref_data, ref_name, name, "name")

    def get_ref_data_tuple(self, ref_data, ref_name, key_value, key_type):
        dict_key = f"{ref_name}{key_type}"
        # Try get the object from the cache
        ref_object = self.ref_data_dicts.get(dict_key, {}).get(key_value.lower(), ())
        if ref_object:
            return ref_object

        # No cache, we need to add it to the cache.
        d = {r[key_type].lower(): (r["id"], r["name"]) for r in ref_data}
        self.ref_data_dicts[dict_key] = d
        return self.ref_data_dicts.get(dict_key, {}).get(key_value.lower(), ())

    def condition_remove_substring(self, legacy_id, value, parameter, marc_field: field.Field):
        return value.replace(parameter["substring"], "")

    def condition_set_instance_type_id(self, legacy_id, value, parameter, marc_field: field.Field):
        if marc_field.tag not in ["008", "336"]:
            self.mapper.migration_report.add(
                "InstanceTypeMapping",
                (
                    f"Unhandled MARC tag {marc_field.tag}. Instance Type ID is only mapped "
                    "from 336 "
                ),
            )
        return ""  # functionality moved

    def condition_set_electronic_access_relations_id(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        return self._extracted_from_condition_set_electronic_access_relations_id_2("3", marc_field)

    # TODO Rename this here and in `condition_set_url_relationship` and `condition_set_electronic_access_relations_id`
    def _extracted_from_condition_set_electronic_access_relations_id_2(self, arg0, marc_field):
        enum = {
            "0": "resource",
            "1": "version of resource",
            "2": "related resource",
            arg0: "no information provided",
        }

        ind2 = marc_field.indicator2
        name = enum.get(ind2, enum[arg0])
        if not self.folio.electronic_access_relationships:
            raise ValueError("No electronic_access_relationships setup in tenant")
        t = self.get_ref_data_tuple_by_name(
            self.folio.electronic_access_relationships,
            "electronic_access_relationships",
            name,
        )

        self.mapper.migration_report.add("MappedElectronicRelationshipTypes", t[1])

        return t[0]

    def condition_set_note_staff_only_via_indicator(
        self, legacy_id, value, parameter, marc_field: field.Field
    ):
        """Returns true of false depending on the first indicator"""
        # https://www.loc.gov/marc/bibliographic/bd541.html
        ind1 = marc_field.indicator1
        self.mapper.migration_report.add(
            "StaffOnlyViaIndicator",
            f"{marc_field.tag} indicator1: {ind1} ("
            + i18n.t("1 is public, all other values are Staff only")
            + ")",
        )
        if ind1 != "1":
            return "true"
        return "false"



import logging
import sys
import time
import traceback
from typing import List
import i18n

from folio_uuid.folio_namespaces import FOLIONamespaces
from pymarc import Field
from pymarc import Record
from pymarc import Subfield

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.folder_structure import FolderStructure
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.marc_rules_transformation.rules_mapper_base import (
    RulesMapperBase,
)
from folio_migration_tools.migration_report import MigrationReport


class MarcFileProcessor:
    def __init__(
        self, mapper: RulesMapperBase, folder_structure: FolderStructure, created_objects_file
    ):
        self.object_type: FOLIONamespaces = folder_structure.object_type
        self.folder_structure: FolderStructure = folder_structure
        self.mapper: RulesMapperBase = mapper
        self.created_objects_file = created_objects_file
        self.srs_records_file = open(self.folder_structure.srs_records_path, "w+")
        self.unique_001s: set = set()
        self.failed_records_count: int = 0
        self.records_count: int = 0
        self.start: float = time.time()
        self.legacy_ids: set = set()
        if (
            self.object_type == FOLIONamespaces.holdings
            and self.mapper.task_configuration.create_source_records
        ):
            logging.info("Loading Parent HRID map for SRS creation")
            self.parent_hrids = {entity[1]: entity[2] for entity in mapper.parent_id_map.values()}

    def process_record(self, idx: int, marc_record: Record, file_def: FileDefinition):
        """processes a marc holdings record and saves it

        Args:
            idx (int): Index in file being parsed
            marc_record (Record): _description_
            file_def (FileDefinition): _description_

        Raises:
            TransformationProcessError: _description_
            TransformationRecordFailedError: _description_
        """
        success = True
        folio_recs = []
        self.records_count += 1
        try:
            # Transform the MARC21 to a FOLIO record
            legacy_ids = self.mapper.get_legacy_ids(marc_record, idx)
            if not legacy_ids:
                raise TransformationRecordFailedError(
                    f"Index in file: {idx}", "No legacy id found", idx
                )
            folio_recs = self.mapper.parse_record(marc_record, file_def, legacy_ids)
            for idx, folio_rec in enumerate(folio_recs):
                if idx == 0:
                    filtered_legacy_ids = self.get_valid_folio_record_ids(
                        legacy_ids, self.legacy_ids, self.mapper.migration_report
                    )
                    self.add_legacy_ids_to_map(folio_rec, filtered_legacy_ids)

                    self.save_srs_record(
                        marc_record,
                        file_def,
                        folio_rec,
                        legacy_ids,
                        self.object_type,
                    )
                Helper.write_to_file(self.created_objects_file, folio_rec)
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Inventory records written to disk")
                )
                self.exit_on_too_many_exceptions()

        except TransformationRecordFailedError as error:
            success = False
            raise TransformationRecordFailedError(
                f"{error.index_or_id} in {file_def.file_name}", error.message, error.data_value
            ) from error
        except TransformationProcessError as tpe:
            raise TransformationProcessError(
                f"{tpe.index_or_id} in {file_def.file_name}", tpe.message, tpe.data_value
            ) from tpe
        except Exception as inst:
            success = False
            traceback.print_exc()
            logging.error(type(inst))
            logging.error(inst.args)
            logging.error(inst)
            logging.error(marc_record)
            logging.error(folio_recs)
            raise TransformationProcessError("", inst.args, "") from inst
        finally:
            if not success:
                self.failed_records_count += 1
                remove_from_id_map = getattr(self.mapper, "remove_from_id_map", None)
                for folio_rec in folio_recs:
                    if (
                        callable(remove_from_id_map)
                        and "folio_rec" in locals()
                        and folio_rec.get("formerIds", "")
                    ):
                        self.mapper.remove_from_id_map(folio_rec.get("formerIds", []))

    def save_srs_record(
        self,
        marc_record: Record,
        file_def: FileDefinition,
        folio_rec,
        legacy_ids: List[str],
        object_type: FOLIONamespaces,
    ):
        if not self.mapper.task_configuration.create_source_records:
            return
        if object_type in [FOLIONamespaces.holdings]:
            if "008" in marc_record and len(marc_record["008"].data) > 32:
                remain, rest = (
                    marc_record["008"].data[:32],
                    marc_record["008"].data[32:],
                )
                marc_record["008"].data = remain
                self.mapper.migration_report.add(
                    "MarcValidation",
                    i18n.t("008 length invalid. '%{rest}' was stripped out", rest=rest),
                )
            self.add_mapped_location_code_to_record(marc_record, folio_rec)
            new_004 = Field(tag="004", data=self.parent_hrids[folio_rec["instanceId"]])
            marc_record.remove_fields("004")
            marc_record.add_ordered_field(new_004)
            for former_id in legacy_ids:
                if self.mapper.task_configuration.hrid_handling == HridHandling.default:
                    new_035 = Field(
                        tag="035",
                        indicators=[" ", " "],
                        subfields=[Subfield(code="a", value=former_id)],
                    )
                    marc_record.add_ordered_field(new_035)
        self.mapper.save_source_record(
            self.srs_records_file,
            self.object_type,
            self.mapper.folio_client,
            marc_record,
            folio_rec,
            legacy_ids,
            file_def.discovery_suppressed,
        )
        self.mapper.migration_report.add_general_statistics(i18n.t("SRS records written to disk"))

    def add_mapped_location_code_to_record(self, marc_record, folio_rec):
        location_code = next(
            (
                location["code"]
                for location in self.mapper.folio_client.locations
                if location["id"] == folio_rec["permanentLocationId"]
            ),
            None,
        )
        if "852" not in marc_record:
            raise TransformationRecordFailedError(
                "", "No 852 in record when storing new location code", ""
            )
        first_852 = marc_record.get_fields("852")[0]
        first_852.delete_subfield("b")
        while old_b := first_852.delete_subfield("b"):
            first_852.add_subfield("x", old_b, 0)
            self.mapper.migration_report.add(
                "LocationMapping", i18n.t("Additional 852$b was moved to 852$x")
            )
        first_852.add_subfield("b", location_code, 0)
        self.mapper.migration_report.add(
            "LocationMapping", i18n.t("Set 852 to FOLIO location code")
        )

    def exit_on_too_many_exceptions(self):
        if (
            self.failed_records_count / (self.records_count + 1)
            > (self.mapper.library_configuration.failed_percentage_threshold / 100)
            and self.failed_records_count
            > self.mapper.library_configuration.failed_records_threshold
        ):
            logging.critical("More than 20 percent of the records have failed. Halting")
            sys.exit(1)

    @staticmethod
    def get_valid_folio_record_ids(
        legacy_ids, folio_record_identifiers, migration_report: MigrationReport
    ):
        new_ids = set()
        for legacy_id in legacy_ids:
            if legacy_id not in folio_record_identifiers:
                new_ids.add(legacy_id)
            else:
                migration_report.add_general_statistics(
                    i18n.t("Duplicate MARC record identifiers ")
                )
        if not any(new_ids):
            s = i18n.t("Failed records. No unique record identifiers in legacy record")
            migration_report.add_general_statistics(s)
            raise TransformationRecordFailedError(
                "-".join(legacy_ids),
                "Duplicate recod identifier(s). See logs. Record Failed",
                "-".join(legacy_ids),
            )
        return list(new_ids)

    def wrap_up(self):
        """Finalizes the mapping by writing things out."""
        logging.info(
            "Saving map of %s old and new IDs to %s",
            len(self.mapper.id_map),
            self.folder_structure.id_map_path,
        )
        self.mapper.save_id_map_file(self.folder_structure.id_map_path, self.mapper.id_map)
        logging.info("%s records processed", self.records_count)
        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("MFHD records transformation report"),
                report_file,
                self.mapper.start_datetime,
            )
            Helper.print_mapping_report(
                report_file,
                self.mapper.parsed_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )
        self.srs_records_file.close()
        self.mapper.wrap_up()

        logging.info("Transformation report written to %s", report_file.name)
        logging.info("Processor is done.")

    def add_legacy_ids_to_map(self, folio_rec, filtered_legacy_ids):
        for legacy_id in filtered_legacy_ids:
            self.legacy_ids.add(legacy_id)
            if legacy_id not in self.mapper.id_map:
                self.mapper.id_map[legacy_id] = self.mapper.get_id_map_tuple(
                    legacy_id, folio_rec, self.object_type
                )

            else:
                raise TransformationRecordFailedError(
                    legacy_id,
                    "Legacy ID already added to Legacy Id map.",
                    ",".join(filtered_legacy_ids),
                )



import json
from pathlib import Path


def deep_merge(target_dict, source_dict, merge_keys=("name", "fileName", "file_name")):
    # deep_merge
    #
    # Deeply merges nested dictionaries and lists of dictionaries.
    # **Muatates the target_dict with the changes**, and returns it.
    #
    # For lists, attempts to match items based on the first found key in merge_keys in each item.
    # If no key is found, or no match, the item is appended to the target list.
    #
    # Deletes any keys with value Null in the source.

    for k in source_dict:
        if isinstance(target_dict.get(k, None), dict) and isinstance(source_dict[k], dict):
            # Recursive, depth-first merge on dictionaries
            target_dict[k] = deep_merge(target_dict[k], source_dict[k], merge_keys)
        elif isinstance(target_dict.get(k, None), list) and isinstance(source_dict[k], list):
            # Merge lists on keys in merge_keys
            for merging_list_item in source_dict[k]:
                if not isinstance(merging_list_item, dict):
                    target_dict[k].append(merging_list_item)
                    continue
                merge_key = next((i for i in merge_keys if i in merging_list_item), None)
                if merge_key is None:
                    target_dict[k].append(merging_list_item)
                    continue
                for index, target in enumerate(target_dict[k]):
                    if target[merge_key] == merging_list_item[merge_key]:
                        target_dict[k][index] = deep_merge(
                            target_dict[k][index],
                            merging_list_item,
                            merge_keys,
                        )
                        break
                else:
                    target_dict[k].append(merging_list_item)
        else:
            if source_dict[k] is None:
                target_dict.pop(k, None)
            else:
                target_dict[k] = source_dict[k]
    return target_dict


def merge_load(config_path_str, parsed_config=None):
    # Recursively load JSON files from a configuration file
    #
    # If a configuration file has a "source" key, either a string or list,
    # the file(s) will be loaded in the order presented, and the config file
    # will be merged on top.
    #
    # To delete a key, set it to `null` in the json source file.

    parsed_config = parsed_config or {}
    config_path = Path(config_path_str)
    with open(config_path) as config_file:
        single_config = json.load(config_file)
    sources = single_config.get("source", [])
    if isinstance(sources, str):
        sources = [sources]
    for source in sources:
        parsed_config = merge_load(config_path.parent / source, parsed_config)
    return deep_merge(parsed_config, single_config)



"""The default mapper, responsible for parsing MARC21 records acording to the
FOLIO community specifications"""
import logging
import re
import time
import uuid
from typing import List

import i18n
import pymarc
from folio_uuid.folio_namespaces import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient
from pymarc import Record

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import IlsFlavour
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.marc_rules_transformation.conditions import Conditions
from folio_migration_tools.marc_rules_transformation.hrid_handler import HRIDHandler
from folio_migration_tools.marc_rules_transformation.rules_mapper_base import (
    RulesMapperBase,
)


class AuthorityMapper(RulesMapperBase):
    non_repatable_fields = [
        "100",
        "110",
        "111",
        "130",
        "147",
        "148",
        "150",
        "151",
        "155",
        "162",
        "180",
        "181",
        "182",
        "185",
        "378",
        "384",
    ]
    """_summary_

    Args:
        RulesMapperBase (_type_): _description_
    """

    def __init__(
        self,
        folio_client,
        library_configuration: LibraryConfiguration,
        task_configuration,
    ):
        super().__init__(
            folio_client,
            library_configuration,
            task_configuration,
            self.get_authority_json_schema(folio_client, library_configuration),
            Conditions(folio_client, self, "auth", library_configuration.folio_release),
        )
        self.srs_recs: list = []
        logging.info("Fetching mapping rules from the tenant")
        rules_endpoint = "/mapping-rules/marc-authority"
        self.mappings = self.folio_client.folio_get_single_object(rules_endpoint)
        self.source_file_mapping: dict = {}
        self.setup_source_file_mapping()
        self.start = time.time()

    def get_legacy_ids(self, marc_record: Record, idx: int) -> List[str]:
        ils_flavour: IlsFlavour = self.task_configuration.ils_flavour
        if ils_flavour in {IlsFlavour.sierra, IlsFlavour.millennium}:
            raise TransformationProcessError("", f"ILS {ils_flavour} not configured")
        elif ils_flavour == IlsFlavour.tag907y:
            return RulesMapperBase.get_bib_id_from_907y(marc_record, idx)
        elif ils_flavour == IlsFlavour.tagf990a:
            return RulesMapperBase.get_bib_id_from_990a(marc_record, idx)
        elif ils_flavour == IlsFlavour.aleph:
            raise TransformationProcessError("", f"ILS {ils_flavour} not configured")
        elif ils_flavour in {IlsFlavour.voyager, "voyager", IlsFlavour.tag001}:
            return RulesMapperBase.get_bib_id_from_001(marc_record, idx)
        elif ils_flavour == IlsFlavour.koha:
            raise TransformationProcessError("", f"ILS {ils_flavour} not configured")
        elif ils_flavour == IlsFlavour.none:
            return [str(uuid.uuid4())]
        else:
            raise TransformationProcessError("", f"ILS {ils_flavour} not configured")

    def parse_record(
        self, marc_record: pymarc.Record, file_def: FileDefinition, legacy_ids: List[str]
    ) -> list[dict]:
        """Parses an auth recod into a FOLIO Authority object
         This is the main function

        Args:
            legacy_ids (_type_): _description_
            marc_record (Record): _description_
            file_def (FileDefinition): _description_

        Returns:
            dict: _description_
        """
        self.print_progress()
        ignored_subsequent_fields: set = set()
        bad_tags = set(self.task_configuration.tags_to_delete)  # "907"
        folio_authority = self.perform_initial_preparation(marc_record, legacy_ids)
        for marc_field in marc_record:
            self.report_marc_stats(marc_field, bad_tags, legacy_ids, ignored_subsequent_fields)
            if marc_field.tag not in ignored_subsequent_fields:
                self.process_marc_field(
                    folio_authority,
                    marc_field,
                    ignored_subsequent_fields,
                    legacy_ids,
                )

        self.perform_additional_parsing(folio_authority)
        clean_folio_authority = self.validate_required_properties(
            "-".join(legacy_ids), folio_authority, self.schema, FOLIONamespaces.instances
        )
        self.dedupe_rec(clean_folio_authority)
        marc_record.remove_fields(*list(bad_tags))
        self.report_folio_mapping(clean_folio_authority, self.schema)
        return [clean_folio_authority]

    def perform_initial_preparation(self, marc_record: pymarc.Record, legacy_ids):
        folio_authority = {}
        folio_authority["id"] = str(
            FolioUUID(
                str(self.folio_client.okapi_url),
                FOLIONamespaces.authorities,
                str(legacy_ids[-1]),
            )
        )
        HRIDHandler.handle_035_generation(
            marc_record, legacy_ids, self.migration_report, False, False
        )
        self.map_source_file_and_natural_id(marc_record, folio_authority)
        self.handle_leader_17(marc_record, legacy_ids)
        return folio_authority

    def map_source_file_and_natural_id(self, marc_record, folio_authority):
        """Implement source file and natural ID mappings according to MODDICORE-283"""
        match_prefix_patt = re.compile("^[A-Za-z]+")
        natural_id = None
        source_file_id = None
        has_010 = marc_record.get("010")
        if has_010 and (has_010a := has_010.get_subfields("a")):
            for a_subfield in has_010a:
                natural_id_prefix = match_prefix_patt.match(a_subfield)
                if natural_id_prefix and (
                    source_file := self.source_file_mapping.get(natural_id_prefix.group(0), None)
                ):
                    natural_id = "".join(a_subfield.split())
                    source_file_id = source_file["id"]
                    self.migration_report.add_general_statistics(
                        i18n.t("naturalId mapped from %{fro}", fro="010$a")
                    )
                    self.migration_report.add(
                        "AuthoritySourceFileMapping",
                        f"{source_file['name']} -- {natural_id_prefix.group(0)} -- 010$a",
                        number=1,
                    )
                    break
        if not source_file_id:
            natural_id = "".join(marc_record["001"].data.split())
            self.migration_report.add_general_statistics(
                i18n.t("naturalId mapped from %{fro}", fro="001")
            )
            natural_id_prefix = match_prefix_patt.match(natural_id)
            if natural_id_prefix:
                if source_file := self.source_file_mapping.get(natural_id_prefix.group(0), None):
                    source_file_id = source_file["id"]
                    self.migration_report.add(
                        "AuthoritySourceFileMapping",
                        f"{source_file['name']} -- {natural_id_prefix.group(0)} -- 001",
                        number=1,
                    )
        folio_authority["naturalId"] = natural_id
        if source_file_id:
            folio_authority["sourceFileId"] = source_file_id

    def setup_source_file_mapping(self):
        if self.folio_client.authority_source_files:
            logging.info(
                f"{len(self.folio_client.authority_source_files)} \tAuthority source files"
            )
            for source_file in self.folio_client.authority_source_files:
                for sf_code in source_file.get("codes", []):
                    self.source_file_mapping[sf_code] = source_file

    def handle_leader_17(self, marc_record, legacy_ids):
        leader_17 = marc_record.leader[17] or "Empty"
        self.migration_report.add(
            "AuthorityEncodingLevel", i18n.t("Original value") + f": {leader_17}"
        )
        if leader_17 not in ["n", "o"]:
            Helper.log_data_issue(
                legacy_ids,
                f"LDR pos. 17 is '{leader_17}'. Is this correct? Value has been changed to 'n'.",
                marc_record.leader,
            )
            marc_record.leader = f"{marc_record.leader[:17]}n{marc_record.leader[18:]}"
            self.migration_report.add(
                "AuthorityEncodingLevel", i18n.t("Changed %{a} to %{b}", a=leader_17, b="n")
            )

    def perform_additional_parsing(
        self,
        folio_authority: dict,
    ) -> None:
        """Do stuff not easily captured by the mapping rules

        Args:
            folio_authority (dict): _description_
        """
        folio_authority["source"] = "MARC"

    def get_authority_json_schema(self, folio_client: FolioClient, library_configuration):
        """Fetches the JSON Schema for autorities"""
        if library_configuration.folio_release.name.lower()[0] < "p":
            schema = folio_client.get_from_github(
                "folio-org", "mod-inventory-storage", "/ramls/authorities/authority.json"
            )
        else:
            schema = folio_client.get_from_github(
                "folio-org",
                "mod-entities-links",
                "/src/main/resources/swagger.api/schemas/authority-storage/authorityDto.yaml",
            )
        return schema

    def wrap_up(self):
        logging.info("Mapper wrapping up")



import copy
import json
import logging
from typing import List

import i18n
from folio_uuid.folio_namespaces import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient
from pymarc.field import Field
from pymarc.record import Record

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.holdings_helper import HoldingsHelper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.marc_rules_transformation.conditions import Conditions
from folio_migration_tools.marc_rules_transformation.holdings_statementsparser import (
    HoldingsStatementsParser,
)
from folio_migration_tools.marc_rules_transformation.rules_mapper_base import (
    RulesMapperBase,
)


class RulesMapperHoldings(RulesMapperBase):
    def __init__(
        self,
        folio_client,
        location_map,
        task_configuration,
        library_configuration: LibraryConfiguration,
        parent_id_map: dict,
        boundwith_relationship_map,
    ):
        self.task_configuration = task_configuration
        self.conditions = Conditions(
            folio_client,
            self,
            "holdings",
            library_configuration.folio_release,
            self.task_configuration.default_call_number_type_name,
        )
        self.folio = folio_client
        super().__init__(
            folio_client,
            library_configuration,
            task_configuration,
            self.fetch_holdings_schema(folio_client),
            self.conditions,
            parent_id_map,
        )
        self.boundwith_relationship_map = self.setup_boundwith_relationship_map(
            boundwith_relationship_map
        )
        self.location_map = location_map
        self.holdings_id_map: dict = {}
        self.ref_data_dicts: dict = {}
        self.fallback_holdings_type_id = self.task_configuration.fallback_holdings_type_id
        self.setup_holdings_sources()
        logging.info("Fetching mapping rules from the tenant")
        rules_endpoint = "/mapping-rules/marc-holdings"
        self.mappings = self.folio_client.folio_get_single_object(rules_endpoint)
        self.fix_853_bug_in_rules()

    def fix_853_bug_in_rules(self):
        f852_mappings = self.mappings["852"]
        new_852_mapping = []
        for mapping in f852_mappings:
            if "entity" in mapping:
                for entity_mapping in mapping["entity"]:
                    if "." not in entity_mapping["target"]:
                        new_852_mapping.append(entity_mapping)
                    else:
                        raise TransformationProcessError(
                            "",
                            (
                                "Actual entity mapping found in 852 mappings. "
                                "Report this to the maintainers of this codebase"
                            ),
                            json.dumps(entity_mapping),
                        )
        self.mappings["852"] = new_852_mapping

    def parse_record(
        self, marc_record: Record, file_def: FileDefinition, legacy_ids: List[str]
    ) -> list[dict]:
        """Parses a mfhd recod into a FOLIO Inventory holdings object
        Community mapping suggestion: https://tinyurl.com/3rh52e2x
         This is the main function

        Args:
            marc_record (Record): _description_
            file_def (FileDefinition): _description_
            legacy_ids (List[str]): _description_

        Raises:
            TransformationRecordFailedError: _description_

        Returns:
            dict: _description_
        """

        self.print_progress()
        folio_holding = self.perform_initial_preparation(marc_record, legacy_ids)

        self.migration_report.add("RecordStatus", marc_record.leader[5])
        ignored_subsequent_fields: set = set()
        num_852s = 0
        for marc_field in marc_record:
            try:
                if marc_field.tag == "852":
                    num_852s += 1
                    if num_852s > 1:
                        continue
                self.process_marc_field(
                    folio_holding,
                    marc_field,
                    ignored_subsequent_fields,
                    legacy_ids,
                )
            except TransformationFieldMappingError as tfme:
                tfme.log_it()
        if num_852s > 1:
            Helper.log_data_issue(legacy_ids, "More than 1 852 found", "")

        self.perform_additional_mapping(marc_record, folio_holding, legacy_ids, file_def)
        cleaned_folio_holding = self.validate_required_properties(
            "-".join(folio_holding.get("formerIds")),
            folio_holding,
            self.schema,
            FOLIONamespaces.holdings,
        )
        if not folio_holding.get("instanceId", ""):
            raise TransformationRecordFailedError(
                legacy_ids,
                "No Instance id mapped. ",
                folio_holding.get("formerIds", ["No former ids"]),
            )
        props_to_not_dedupe = (
            []
            if self.task_configuration.deduplicate_holdings_statements
            else [
                "holdingsStatements",
                "holdingsStatementsForIndexes",
                "holdingsStatementsForSupplements",
            ]
        )
        self.dedupe_rec(cleaned_folio_holding, props_to_not_dedupe)
        self.report_folio_mapping(cleaned_folio_holding, self.schema)
        if bw_instance_ids := self.boundwith_relationship_map.get(cleaned_folio_holding["id"], []):
            return list(
                self.create_bound_with_holdings(
                    cleaned_folio_holding,
                    bw_instance_ids,
                    self.task_configuration.holdings_type_uuid_for_boundwiths,
                )
            )
        return [cleaned_folio_holding]

    def set_instance_id_by_map(self, legacy_ids: list, folio_holding: dict, marc_record: Record):
        if "004" not in marc_record:
            raise TransformationProcessError(
                "",
                ("No 004 in record. The tools only support bib-mfhd linking throuh 004"),
                legacy_ids,
            )
        legacy_instance_id = marc_record["004"].data.strip()
        folio_holding["formerIds"].append(f"{self.bib_id_template}{legacy_instance_id}")
        if legacy_instance_id in self.parent_id_map:
            folio_holding["instanceId"] = self.parent_id_map[legacy_instance_id][1]
        else:
            raise TransformationRecordFailedError(
                legacy_ids,
                "Old instance id not in map",
                marc_record["004"],
            )

    def perform_initial_preparation(self, marc_record: Record, legacy_ids):
        folio_holding: dict = {}
        folio_holding["id"] = str(
            FolioUUID(
                str(self.folio_client.okapi_url),
                FOLIONamespaces.holdings,
                str(legacy_ids[0]),
            )
        )
        for legacy_id in legacy_ids:
            self.add_legacy_id_to_admin_note(folio_holding, legacy_id)
        folio_holding["formerIds"] = copy.copy(legacy_ids)
        self.set_instance_id_by_map(legacy_ids, folio_holding, marc_record)
        return folio_holding

    def setup_holdings_sources(self):
        holdings_sources = list(
            self.folio_client.folio_get_all("/holdings-sources", "holdingsRecordsSources")
        )
        logging.info("Fetched %s holdingsRecordsSources from tenant", len(holdings_sources))
        self.holdingssources = {n["name"].upper(): n["id"] for n in holdings_sources}
        if "FOLIO" not in self.holdingssources:
            raise TransformationProcessError("", "No holdings source with name FOLIO in tenant")
        if "MARC" not in self.holdingssources:
            raise TransformationProcessError("", "No holdings source with name MARC in tenant")

    def process_marc_field(
        self,
        folio_holding: dict,
        marc_field: Field,
        ignored_subsequent_fields,
        index_or_legacy_ids,
    ):
        """This overwrites the implementation for Auth and instances

        Args:
            folio_holding (dict): _description_
            marc_field (Field): _description_
            ignored_subsequent_fields (_type_): _description_
            index_or_legacy_ids (_type_): _description_
        """
        self.migration_report.add("Trivia", i18n.t("Total number of Tags processed"))
        if marc_field.tag not in self.mappings:
            self.report_legacy_mapping(marc_field.tag, True, False)
        elif marc_field.tag not in ignored_subsequent_fields:
            mappings = self.mappings[marc_field.tag]
            self.map_field_according_to_mapping(
                marc_field, mappings, folio_holding, index_or_legacy_ids
            )
            self.report_legacy_mapping(marc_field.tag, True, True)
            if any(m.get("ignoreSubsequentFields", False) for m in mappings):
                ignored_subsequent_fields.add(marc_field.tag)

    def perform_additional_mapping(
        self, marc_record: Record, folio_holding, legacy_ids: List[str], file_def: FileDefinition
    ):
        """_summary_

        Args:
            marc_record (Record): _description_
            folio_holding (_type_): _description_
            legacy_ids (List[str]): _description_
            file_def (FileDefinition): _description_

        Raises:
            TransformationRecordFailedError: _description_
        """
        self.set_holdings_type(marc_record, folio_holding, legacy_ids)
        self.set_default_call_number_type_if_empty(folio_holding)
        self.pick_first_location_if_many(folio_holding, legacy_ids)
        self.parse_coded_holdings_statements(marc_record, folio_holding, legacy_ids)
        HoldingsHelper.handle_notes(folio_holding)
        if (
            self.task_configuration.create_source_records
            or self.task_configuration.hrid_handling == HridHandling.preserve001
        ):
            self.hrid_handler.handle_hrid(
                FOLIONamespaces.holdings, folio_holding, marc_record, legacy_ids
            )
        else:
            del folio_holding["hrid"]
        if not folio_holding.get("instanceId", ""):
            raise TransformationRecordFailedError(
                "".join(folio_holding.get("formerIds", [])),
                "Missing instance ids. Something is wrong.",
                "",
            )
        self.handle_suppression(folio_holding, file_def, True)
        self.set_source_id(self.task_configuration, folio_holding, self.holdingssources)

    def pick_first_location_if_many(self, folio_holding, legacy_ids: List[str]):
        if " " in folio_holding.get("permanentLocationId", ""):
            Helper.log_data_issue(
                legacy_ids,
                "Space in permanentLocationId. Was this MFHD attached to multiple holdings?",
                folio_holding["permanentLocationId"],
            )
            folio_holding["permanentLocationId"] = folio_holding["permanentLocationId"].split(" ")[
                0
            ]

    @staticmethod
    def set_source_id(task_configuration, folio_rec, holdingssources):
        if task_configuration.create_source_records:
            folio_rec["sourceId"] = holdingssources.get("MARC")
        else:
            folio_rec["sourceId"] = holdingssources.get("FOLIO")

    def parse_coded_holdings_statements(
        self, marc_record: Record, folio_holding, legacy_ids: List[str]
    ):
        # TODO: Should one be able to switch these things off?
        a = {
            "holdingsStatements": ("853", "863", "866"),
            "holdingsStatementsForIndexes": ("855", "865", "868"),
            "holdingsStatementsForSupplements": ("854", "864", "867"),
        }
        for key, v in a.items():
            try:
                res = HoldingsStatementsParser.get_holdings_statements(
                    marc_record,
                    v[0],
                    v[1],
                    v[2],
                    legacy_ids,
                    self.task_configuration.deduplicate_holdings_statements,
                )
                if res["statements"]:
                    folio_holding[key] = res["statements"]
                for mr in res["migration_report"]:
                    self.migration_report.add("HoldingsStatementsParsing", f"{mr[0]} -- {mr[1]}")
            except TransformationFieldMappingError as tfme:
                Helper.log_data_issue(tfme.index_or_id, tfme.message, tfme.data_value)
                self.migration_report.add("FieldMappingErrors", tfme.message)

    def wrap_up(self):
        logging.info("Mapper wrapping up")
        if self.task_configuration.update_hrid_settings:
            if self.task_configuration.create_source_records:
                self.hrid_handler.store_hrid_settings()
            else:
                logging.info("NOT storing HRID settings since that is managed by FOLIO")

    def fetch_holdings_schema(self, folio_client: FolioClient):
        logging.info("Fetching HoldingsRecord schema...")
        return folio_client.get_from_github(
            "folio-org", "mod-inventory-storage", "ramls/holdingsrecord.json"
        )

    def set_holdings_type(self, marc_record: Record, folio_holding, legacy_ids: List[str]):
        # Holdings type mapping
        ldr06 = marc_record.leader[6]
        # TODO: map this better
        # type = type_map.get(ldr06, "Unknown")
        if folio_holding.get("holdingsTypeId", ""):
            self.migration_report.add(
                "HoldingsTypeMapping",
                i18n.t(
                    "Already set to %{value}. %{leader_key} was %{leader}",
                    value=folio_holding.get("holdingsTypeId"),
                    leader_key="LDR[06]",
                    leader=ldr06,
                ),
            )
        else:
            holdings_type = self.conditions.holdings_type_map.get(ldr06, "")
            if t := self.conditions.get_ref_data_tuple_by_name(
                self.conditions.holdings_types, "hold_types", holdings_type
            ):
                folio_holding["holdingsTypeId"] = t[0]
                self.migration_report.add(
                    "HoldingsTypeMapping",
                    f"{ldr06} -> {holdings_type} -> {t[1]} ({t[0]}",
                )
                if holdings_type == "Unknown":
                    Helper.log_data_issue(
                        legacy_ids,
                        (
                            i18n.t("blurbs.HoldingsTypeMapping.title") + " is 'unknown'. "
                            "(leader 06 is set to 'u') Check if this is correct"
                        ),
                        ldr06,
                    )
            else:
                if not self.fallback_holdings_type_id:
                    raise TransformationProcessError(
                        "",
                        "No fallbackHoldingsTypeId set up. Add to task configuration",
                    )
                folio_holding["holdingsTypeId"] = self.fallback_holdings_type_id
                self.migration_report.add(
                    "HoldingsTypeMapping",
                    i18n.t("An Unmapped")
                    + f" {ldr06} -> {holdings_type} -> "
                    + i18n.t("Unmapped"),
                )
                Helper.log_data_issue(
                    legacy_ids,
                    (
                        i18n.t("blurbs.HoldingsTypeMapping.title", locale="en")
                        + ". leader 06 was unmapped."
                    ),
                    ldr06,
                )

    def set_default_call_number_type_if_empty(self, folio_holding):
        if not folio_holding.get("callNumberTypeId", ""):
            folio_holding["callNumberTypeId"] = self.conditions.default_call_number_type["id"]

    def get_legacy_ids(self, marc_record: Record, idx: int) -> List[str]:
        marc_path = self.task_configuration.legacy_id_marc_path
        split = marc_path.split("$", maxsplit=1)
        results = []
        if not (split[0].isnumeric() and len(split[0]) == 3):
            raise TransformationProcessError(
                "",
                (
                    "the marc field used for determining the legacy id is not numeric "
                    "or does not have the stipulated lenght of 3."
                    "Make sure the task configuration setting for 'legacyIdMarcPath' "
                    "is correct or make this piece of code more allowing"
                ),
                marc_path,
            )
        elif len(split) == 1:
            results.append(marc_record[split[0]].value())
        elif len(split) == 2 and len(split[1]) == 1:
            for field in marc_record.get_fields(split[0]):
                if sf := field.get_subfields(split[1]):
                    results.append(sf[0])
        else:
            raise TransformationProcessError(
                "",
                ("Something is wrong with 'legacyIdMarcPath' property in the settings"),
                marc_path,
            )
        if not any(results):
            raise TransformationRecordFailedError(
                idx, f"No legacy id found in record from {marc_path}", ""
            )
        return results



import logging
import sys
import i18n
from io import IOBase
from pathlib import Path

from pymarc import MARCReader
from pymarc import Record

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.folder_structure import FolderStructure
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.migration_report import MigrationReport


class MARCReaderWrapper:
    @staticmethod
    def process_single_file(
        file_def: FileDefinition,
        processor,
        failed_records_path: Path,
        folder_structure: FolderStructure,
    ):
        try:
            with open(failed_records_path, "ab") as failed_marc_records_file:
                with open(
                    folder_structure.legacy_records_folder / file_def.file_name,
                    "rb",
                ) as marc_file:
                    reader = MARCReader(marc_file, to_unicode=True, permissive=True)
                    reader.hide_utf8_warnings = True
                    reader.force_utf8 = False
                    logging.info("Running %s", file_def.file_name)
                    MARCReaderWrapper.read_records(
                        reader, file_def, failed_marc_records_file, processor
                    )
        except TransformationProcessError as tpe:
            logging.critical(tpe)
            sys.exit(1)
        except Exception:
            logging.exception("Failure in Main: %s", file_def.file_name, stack_info=True)

    @staticmethod
    def read_records(
        reader,
        source_file: FileDefinition,
        failed_records_file: IOBase,
        processor,
    ):
        for idx, record in enumerate(reader):
            processor.mapper.migration_report.add_general_statistics(
                i18n.t("Records in file before parsing")
            )
            try:
                # None = Something bad happened
                if record is None:
                    report_failed_parsing(
                        reader,
                        source_file,
                        failed_records_file,
                        idx,
                        processor.mapper.migration_report,
                    )
                # The normal case
                else:
                    MARCReaderWrapper.set_leader(record, processor.mapper.migration_report)
                    processor.mapper.migration_report.add_general_statistics(
                        i18n.t("Records successfully decoded from MARC21"),
                    )
                    processor.process_record(idx, record, source_file)
            except TransformationRecordFailedError as error:
                error.log_it()
                processor.mapper.migration_report.add_general_statistics(
                    i18n.t("Records that failed transformation. Check log for details"),
                )
            except ValueError as error:
                logging.error(error)
        logging.info("Done reading %s records from file", idx + 1)

    @staticmethod
    def set_leader(marc_record: Record, migration_report: MigrationReport):
        if marc_record.leader[9] != "a":
            migration_report.add(
                "LeaderManipulation",
                i18n.t(
                    "Set leader 09 (Character coding scheme) from %{field} to a",
                    field=marc_record.leader[9],
                ),
            )
            marc_record.leader = f"{marc_record.leader[:9]}a{marc_record.leader[10:]}"

        if not marc_record.leader.endswith("4500"):
            migration_report.add(
                "LeaderManipulation",
                i18n.t("Set leader 20-23 from %{field} to 4500", field=marc_record.leader[-4:]),
            )
            marc_record.leader = f"{marc_record.leader[:-4]}4500"

        if marc_record.leader[10] != "2":
            migration_report.add(
                "LeaderManipulation",
                i18n.t(
                    "Set leader 10 (Indicator count) from %{field} to 2",
                    field=marc_record.leader[10],
                ),
            )
            marc_record.leader = f"{marc_record.leader[:10]}2{marc_record.leader[11:]}"

        if marc_record.leader[11] != "2":
            migration_report.add(
                "LeaderManipulation",
                i18n.t(
                    "Set leader 11 (Subfield code count) from %{record} to 2",
                    record=marc_record.leader[11],
                ),
            )
            marc_record.leader = f"{marc_record.leader[:11]}2{marc_record.leader[12:]}"


def report_failed_parsing(
    reader, source_file, failed_bibs_file, idx, migration_report: MigrationReport
):
    migration_report.add_general_statistics(
        i18n.t("Records with encoding errors - parsing failed"),
    )
    failed_bibs_file.write(reader.current_chunk)
    raise TransformationRecordFailedError(
        f"Index in {source_file.file_name}:{idx}",
        f"MARC parsing error: {reader.current_exception}",
        "Failed records stored in results/failed_bib_records.mrc",
    )



import csv
import json
import logging
import os
import sys
import time
from abc import abstractmethod
from datetime import datetime
from datetime import timezone
from pathlib import Path

from folio_uuid.folio_namespaces import FOLIONamespaces
from folioclient import FolioClient
from genericpath import isfile

from folio_migration_tools import library_configuration
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.extradata_writer import ExtradataWriter
from folio_migration_tools.folder_structure import FolderStructure
from folio_migration_tools.marc_rules_transformation.marc_file_processor import (
    MarcFileProcessor,
)
from folio_migration_tools.marc_rules_transformation.marc_reader_wrapper import (
    MARCReaderWrapper,
)


class MigrationTaskBase:
    @staticmethod
    @abstractmethod
    def get_object_type() -> FOLIONamespaces:
        raise NotImplementedError()

    def __init__(
        self,
        library_configuration: library_configuration.LibraryConfiguration,
        task_configuration,
        use_logging: bool = True,
    ):
        logging.info("MigrationTaskBase init")
        self.start_datetime = datetime.now(timezone.utc)
        self.task_configuration = task_configuration
        logging.info(self.task_configuration.json(indent=4))
        self.folio_client: FolioClient = FolioClient(
            library_configuration.okapi_url,
            library_configuration.tenant_id,
            library_configuration.okapi_username,
            library_configuration.okapi_password,
        )
        self.folder_structure: FolderStructure = FolderStructure(
            library_configuration.base_folder,
            self.get_object_type(),
            task_configuration.name,
            library_configuration.iteration_identifier,
            library_configuration.add_time_stamp_to_file_names,
        )

        self.library_configuration = library_configuration
        self.object_type = self.get_object_type()
        try:
            self.folder_structure.setup_migration_file_structure()
            # Initiate Worker
        except FileNotFoundError as fne:
            logging.error(fne)
        except TransformationProcessError as process_error:
            logging.critical(process_error)
            logging.critical("Halting...")
            sys.exit(1)
        self.num_exeptions: int = 0
        self.extradata_writer = ExtradataWriter(
            self.folder_structure.transformation_extra_data_path
        )
        if use_logging:
            self.setup_logging()
        self.folder_structure.log_folder_structure()
        logging.info("MigrationTaskBase init done")

    @abstractmethod
    def wrap_up(self):
        raise NotImplementedError()

    def clean_out_empty_logs(self):
        if (
            self.folder_structure.data_issue_file_path.is_file()
            and os.stat(self.folder_structure.data_issue_file_path).st_size == 0
        ):
            logging.info("Removing data issues file since it is empty")
            os.remove(self.folder_structure.data_issue_file_path)
            logging.info("Removed data issues file since it was empty")

        if (
            self.folder_structure.failed_marc_recs_file.is_file()
            and os.stat(self.folder_structure.failed_marc_recs_file).st_size == 0
        ):
            os.remove(self.folder_structure.failed_marc_recs_file)
            logging.info("Removed empty failed marc records file since it was empty")

    @abstractmethod
    def do_work(self):
        raise NotImplementedError

    @staticmethod
    def check_source_files(
        source_path: Path, file_defs: list[library_configuration.FileDefinition]
    ) -> None:
        """Lists the source data files. Special case since we use the Items folder for holdings

        Args:
            source_path (Path): _description_
            file_defs (list[library_configuration.FileDefinition]): _description_

        Raises:
            TransformationProcessError: _description_

        """
        files = [source_path / f.file_name for f in file_defs if isfile(source_path / f.file_name)]
        ret_str = ", ".join(f.file_name for f in file_defs)

        if files and len(files) < len(file_defs):
            raise TransformationProcessError(
                "",
                f"Some files listed in task configuration not found in {source_path}."
                f"Listed files: {ret_str}",
            )
        if not any(files):
            raise TransformationProcessError(
                "",
                f"None of the files listed in task configuration found in {source_path}."
                f"Listed files: {ret_str}",
            )
        logging.info("Files to process:")
        for filename in files:
            logging.info("\t%s", filename)

    @staticmethod
    def load_id_map(map_path, raise_if_empty=False):
        if not isfile(map_path):
            logging.warn("No legacy id map found at %s. Will build one from scratch", map_path)
            return {}
        id_map = {}
        loaded_rows = 0
        with open(map_path) as id_map_file:
            for index, json_string in enumerate(id_map_file, start=1):
                loaded_rows = index
                # {"legacy_id", "folio_id","suppressed"}
                map_tuple = json.loads(json_string)
                if loaded_rows % 500000 == 0:
                    print(
                        f"{loaded_rows + 1} ids loaded to map. Last Id: {map_tuple[0]}",
                        end="\r",
                    )

                id_map[map_tuple[0]] = map_tuple
        logging.info("Loaded %s migrated IDs", loaded_rows)
        if not any(id_map) and raise_if_empty:
            raise TransformationProcessError("", "Legacy id map is empty", map_path)
        return id_map

    @staticmethod
    def add_argument(parser, destination, help, **kwargs):
        parser.add_argument(dest=destination, help=help, **kwargs)

    def setup_logging(self):
        debug = self.library_configuration.log_level_debug

        DATA_ISSUE_LVL_NUM = 26
        logging.addLevelName(DATA_ISSUE_LVL_NUM, "DATA_ISSUES")

        def data_issues(self, message, *args, **kws):
            if self.isEnabledFor(DATA_ISSUE_LVL_NUM):
                # Yes, logger takes its '*args' as 'args'.
                self._log(DATA_ISSUE_LVL_NUM, message, args, **kws)

        logging.Logger.data_issues = data_issues
        logger = logging.getLogger()
        logger.propogate = True
        logger.handlers = []
        formatter = logging.Formatter(
            "%(asctime)s\t%(levelname)s\t%(message)s\t%(task_configuration_name)s"
        )
        stream_handler = logging.StreamHandler()
        stream_handler.addFilter(ExcludeLevelFilter(26))
        stream_handler.addFilter(TaskNameFilter(self.task_configuration.name))
        if debug:
            logger.setLevel(logging.DEBUG)
            stream_handler.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
            stream_handler.setLevel(logging.INFO)
            stream_handler.addFilter(ExcludeLevelFilter(30))  # Loose warnings from pymarc
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

        file_formatter = logging.Formatter(
            "%(asctime)s\t%(message)s\t%(task_configuration_name)s\t%(filename)s:%(lineno)d"
        )
        file_handler = logging.FileHandler(
            filename=self.folder_structure.transformation_log_path, mode="w"
        )
        file_handler.addFilter(ExcludeLevelFilter(26))
        file_handler.addFilter(TaskNameFilter(self.task_configuration.name))
        # file_handler.addFilter(LevelFilter(0, 20))
        file_handler.setFormatter(file_formatter)
        file_handler.setLevel(logging.INFO)
        logging.getLogger().addHandler(file_handler)

        # Data issue file formatter
        data_issue_file_formatter = logging.Formatter("%(message)s")
        data_issue_file_handler = logging.FileHandler(
            filename=str(self.folder_structure.data_issue_file_path), mode="w"
        )
        data_issue_file_handler.addFilter(LevelFilter(26))
        data_issue_file_handler.setFormatter(data_issue_file_formatter)
        data_issue_file_handler.setLevel(26)
        logging.getLogger().addHandler(data_issue_file_handler)
        logger.info("Logging set up")

    def setup_records_map(self, mapping_file_path):
        with open(mapping_file_path) as mapping_file:
            field_map = json.load(mapping_file)
            logging.info("%s fields present in record mapping file", len(field_map["data"]))
            mapped_fields = (
                f
                for f in field_map["data"]
                if f["legacy_field"] and f["legacy_field"] != "Not mapped"
            )
            logging.info("%s fields mapped in record mapping file", len(list(mapped_fields)))
            return field_map

    def log_and_exit_if_too_many_errors(self, error: TransformationRecordFailedError, idx):
        self.num_exeptions += 1
        error.log_it()
        if self.num_exeptions / (1 + idx) > 0.2 and self.num_exeptions > 5000:
            logging.fatal(
                f"Number of exceptions exceeded limit of "
                f"{self.num_exeptions}. and failure rate is "
                f"{self.num_exeptions / (1 + idx)} Stopping."
            )
            sys.exit(1)

    @staticmethod
    def print_progress(num_processed, start_time):
        if num_processed > 1 and num_processed % 10000 == 0:
            elapsed = num_processed / (time.time() - start_time)
            elapsed_formatted = "{0:.4g}".format(elapsed)
            logging.info(f"{num_processed:,} records processed. Recs/sec: {elapsed_formatted} ")

    def do_work_marc_transformer(
        self,
    ):
        logging.info("Starting....")
        if self.folder_structure.failed_marc_recs_file.is_file():
            os.remove(self.folder_structure.failed_marc_recs_file)
            logging.info("Removed failed marc records file to prevent duplicating data")
        with open(self.folder_structure.created_objects_path, "w+") as created_records_file:
            self.processor = MarcFileProcessor(
                self.mapper, self.folder_structure, created_records_file
            )
            for file_def in self.task_configuration.files:
                MARCReaderWrapper.process_single_file(
                    file_def,
                    self.processor,
                    self.folder_structure.failed_marc_recs_file,
                    self.folder_structure,
                )

    def load_ref_data_mapping_file(
        self,
        folio_property_name: str,
        map_file_path: Path,
        folio_keys,
        required: bool = True,
    ):
        if (
            folio_property_name in folio_keys
            or required
            or folio_property_name.startswith("statisticalCodeIds")
            or folio_property_name.startswith("locationMap")
        ):
            try:
                with open(map_file_path) as map_file:
                    ref_data_map = list(csv.DictReader(map_file, dialect="tsv"))
                    if not ref_data_map:
                        raise TransformationProcessError("", f"Map has no rows: {map_file_path}")
                    logging.info(
                        "Found %s rows in %s map",
                        len(ref_data_map),
                        folio_property_name,
                    )
                    if not any(ref_data_map[0].keys()):
                        raise TransformationProcessError(
                            "",
                            (
                                f"{folio_property_name} not mapped in legacy->folio mapping file "
                                f"({map_file_path}). Did you map this field, "
                                "but forgot to add a mapping file?"
                            ),
                        )
                    logging.info(
                        "%s will be used for determinig %s",
                        ", ".join(ref_data_map[0].keys()),
                        folio_property_name,
                    )
                    return ref_data_map
            except Exception as exception:
                raise exception

        else:
            logging.info("No mapping setup for %s", folio_property_name)
            logging.info("%s will have default mapping if any ", folio_property_name)
            logging.info(
                "Add a file named %s and add the field to the item.mapping.json file.",
                map_file_path,
            )
            return None


class ExcludeLevelFilter(logging.Filter):
    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record):
        return record.levelno != self.level


class TaskNameFilter(logging.Filter):
    def __init__(self, task_configuration_name):
        super().__init__()
        self.task_configuration_name = task_configuration_name

    def filter(self, record):
        record.task_configuration_name = self.task_configuration_name
        return True


class LevelFilter(logging.Filter):
    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record):
        return record.levelno == self.level



import json
import logging
import i18n
from uuid import uuid4

from folio_migration_tools import custom_exceptions
from folio_migration_tools import helper
from folio_migration_tools.migration_report import MigrationReport


class HoldingsHelper:
    @staticmethod
    def to_key(
        holdings_record: dict,
        fields_criterias: list[str],
        migration_report: MigrationReport,
        holdings_type_id_to_exclude_from_merging: str = "Not set",
    ) -> str:
        """Creates a key from values determined by the fields_crieteria in a holding
        record to determine uniquenes

        fields_criterias are limited to the strings and UUID properties on the first level of
        the object. If the property is not found, or empty, it will be ignored

        IF the holdings type id is matched to holdings_type_id_to_exclude_from_merging,
        the key will be added with a uuid to prevent merging of this holding

        Args:
            holdings_record (dict): The Holdingsrecord
            fields_criterias (list[str]): names of the properties of the holdingsrecord.
            migration_report (MigrationReport): Report to help reporting merge
            holdings_type_id_to_exclude_from_merging: (str): the holdings type UUID to exclude

        Raises:
            Exception: _description_

        Returns:
            str: _description_
        """
        try:
            values = []
            for fields_criteria in fields_criterias:
                v = holdings_record.get(fields_criteria, "")
                if not v:
                    migration_report.add(
                        "HoldingsMerging",
                        i18n.t(
                            "%{fields_criteria} empty or not set", fields_criteria=fields_criteria
                        ),
                    )
                values.append(v)

            if holdings_record.get("holdingsTypeId") == holdings_type_id_to_exclude_from_merging:
                values.append(str(uuid4()))
                migration_report.add(
                    "HoldingsMerging",
                    i18n.t("Holding prevented from merging by holdingsTypeId"),
                )
            return "-".join(values)
        except Exception as exception:
            logging.error(json.dumps(holdings_record, indent=4))
            raise exception from exception

    @staticmethod
    def load_previously_generated_holdings(
        holdings_file_path,
        fields_criteria,
        migration_report: MigrationReport,
        holdings_type_id_to_exclude_from_merging: str = "Not set",
    ):
        if not holdings_file_path.is_file():
            raise custom_exceptions.TransformationProcessError(
                "", "File not found", holdings_file_path
            )
        logging.info(
            "Holdings type id to exclude is set to %s",
            holdings_type_id_to_exclude_from_merging,
        )
        with open(holdings_file_path) as holdings_file:
            prev_holdings = {}
            for row in holdings_file:
                stored_holding = json.loads(row.split("\t")[-1])
                stored_key = HoldingsHelper.to_key(
                    stored_holding,
                    fields_criteria,
                    migration_report,
                    holdings_type_id_to_exclude_from_merging,
                )
                if stored_key in prev_holdings:
                    message = (
                        f"Previously stored holdings key already exists in the list of previously"
                        f" stored Holdings. You have likely not used the same matching criterias"
                        f" ({fields_criteria}) as you did in the previous process"
                    )
                    helper.Helper.log_data_issue(stored_holding["formerIds"], message, stored_key)
                    logging.warn(message)
                    prev_holdings[stored_key] = HoldingsHelper.merge_holding(
                        prev_holdings[stored_key], stored_holding
                    )
                    migration_report.add(
                        "HoldingsMerging",
                        i18n.t("Duplicate key based on current merge criteria. Records merged"),
                    )
                else:
                    migration_report.add(
                        "HoldingsMerging",
                        i18n.t("Previously transformed holdings record loaded"),
                    )
                    prev_holdings[stored_key] = stored_holding
            return prev_holdings

    @staticmethod
    def merge_holding(holdings_record: dict, incoming_holdings: dict) -> dict:
        extend_list("holdingsStatementsForIndexes", holdings_record, incoming_holdings, True)
        extend_list("holdingsStatements", holdings_record, incoming_holdings, True)
        extend_list("holdingsStatementsForSupplements", holdings_record, incoming_holdings, True)
        extend_list("notes", holdings_record, incoming_holdings)
        holdings_record["notes"] = dedupe(holdings_record.get("notes", []))
        extend_list("formerIds", holdings_record, incoming_holdings)
        extend_list("electronicAccess", holdings_record, incoming_holdings)
        HoldingsHelper.remove_empty_holdings_statements(holdings_record)
        merge_boolean("discoverySuppress", holdings_record, incoming_holdings)
        return holdings_record

    @staticmethod
    def remove_empty_holdings_statements(holdings_record: dict):
        keys = [
            "holdingsStatements",
            "holdingsStatementsForIndexes",
            "holdingsStatementsForSupplements",
        ]

        for key in keys:
            if key in holdings_record:
                temp_l = [stmt for stmt in holdings_record[key] if any(stmt.values())]
                holdings_record[key] = temp_l
            if key in holdings_record and not holdings_record.get(key, []):
                del holdings_record[key]

    @staticmethod
    def handle_notes(folio_object):
        if folio_object.get("notes", []):
            filtered_notes = []
            for note_obj in folio_object.get("notes", []):
                if not note_obj.get("holdingsNoteTypeId", ""):
                    raise custom_exceptions.TransformationProcessError(
                        folio_object.get("legacyIds", ""),
                        "Missing note type id mapping",
                        json.dumps(note_obj),
                    )
                elif note_obj.get("note", ""):
                    filtered_notes.append(note_obj)
            if filtered_notes:
                folio_object["notes"] = filtered_notes
            else:
                del folio_object["notes"]


def extend_list(
    prop_name: str, holdings_record: dict, incoming_holdings: dict, accept_dupe_items: bool = False
):
    temp = holdings_record.get(prop_name, [])
    all_already_in = all(i in temp for i in incoming_holdings.get(prop_name, []))
    for f in incoming_holdings.get(prop_name, []):
        if not all_already_in and (accept_dupe_items or f not in temp):
            temp.append(f)
    if temp:
        holdings_record[prop_name] = temp


def dedupe(list_of_dicts):
    return [dict(t) for t in {tuple(d.items()) for d in list_of_dicts}]


def merge_boolean(prop_name: str, holdings_record: dict, incoming_holdings: dict):
    if (
        holdings_record.get(prop_name, False) is True
        and incoming_holdings.get(prop_name, False) is True
    ):
        holdings_record[prop_name] = True
    else:
        holdings_record[prop_name] = False



import json
import logging
import sys

from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationProcessError


class RefDataMapping(object):
    def __init__(
        self,
        folio_client: FolioClient,
        ref_data_path,
        array_name,
        the_map,
        key_type,
        blurb_id,
    ):
        self.name = array_name
        self.cache: dict = {}
        self.blurb_id = blurb_id
        logging.info("%s reference data mapping. Initializing", self.name)
        logging.info("Fetching %s reference data from FOLIO", self.name)
        self.ref_data = list(folio_client.folio_get_all(ref_data_path, array_name, "", 1000))
        self.map = the_map
        self.regular_mappings: list = []
        self.key_type = key_type
        self.hybrid_mappings = []
        self.mapped_legacy_keys = []
        self.default_id = ""
        self.default_name = ""
        self.cached_dict = {}
        self.setup_mappings()
        logging.info("%s reference data mapping. Done init", self.name)

    def get_ref_data_tuple(self, key_value):
        if ref_object := self.cached_dict.get(key_value.lower().strip(), ()):
            return ref_object
        self.cached_dict = {
            r[self.key_type].lower(): (r["id"], r[self.key_type]) for r in self.ref_data
        }
        return self.cached_dict.get(key_value.lower().strip(), ())

    def setup_mappings(self):
        self.pre_validate_map()
        for idx, mapping in enumerate(self.map):
            try:
                # Get the legacy keys
                if idx == 0:
                    self.mapped_legacy_keys = get_mapped_legacy_keys(mapping)
                if self.is_default_mapping(mapping):
                    if t := self.get_ref_data_tuple(mapping[f"folio_{self.key_type}"]):
                        self.default_id = t[0]
                        self.default_name = t[1]
                        logging.info(
                            "Set %s as default %s mapping",
                            mapping[f"folio_{self.key_type}"],
                            self.name,
                        )
                    else:
                        x = mapping.get(f"folio_{self.key_type}", "")
                        raise TransformationProcessError(
                            "",
                            f"No {self.name} - {x} - set up in map or tenant. Check for "
                            f"inconstencies in {self.name} naming. "
                            f"Add a row to mapping file with *:s and a valid {self.name}",
                        )
                else:
                    if self.is_hybrid_default_mapping(mapping):
                        self.hybrid_mappings.append(mapping)
                    else:
                        self.regular_mappings.append(mapping)
                    t = self.get_ref_data_tuple(mapping[f"folio_{self.key_type}"])
                    if not t:
                        raise TransformationProcessError("", f"Mapping not found for {mapping}")
                    mapping["folio_id"] = t[0]
            except TransformationProcessError as transformation_process_error:
                raise transformation_process_error from transformation_process_error
            except Exception as ee:
                logging.info(json.dumps(self.map, indent=4))
                logging.error(ee)
                raise TransformationProcessError(
                    "",
                    f'"{mapping[f"folio_{self.key_type}"]}" could not be found in FOLIO',
                ) from ee

        self.post_validate_map()
        logging.info(
            f"Loaded {len(self.regular_mappings)} mappings for {len(self.ref_data)} {self.name} "
            "in FOLIO"
        )
        logging.info(
            f"loaded {len(self.hybrid_mappings)} hybrid mappings for {len(self.ref_data)} "
            f"{self.name} in FOLIO"
        )

    def get_hybrid_mapping(self, legacy_object):
        obj_key = "_".join(legacy_object[k].strip() for k in self.mapped_legacy_keys)
        if obj_key in self.cache:
            return self.cache[obj_key]
        highest_match = None
        highest_match_number = 0

        prepped_props = {k: legacy_object[k].strip() for k in self.mapped_legacy_keys}
        for mapping in self.hybrid_mappings:
            mismatch = 0
            match_numbers = []
            for k in self.mapped_legacy_keys:
                if mapping[k] == prepped_props[k]:
                    match_numbers.append(10)
                elif mapping[k] == "*":
                    match_numbers.append(1)
                else:
                    mismatch += 1
            summa = sum(match_numbers)

            if mismatch < 1 and summa > highest_match_number and min(match_numbers) > 0:
                highest_match_number = summa
                highest_match = mapping
        self.cache[obj_key] = highest_match
        return highest_match

    def get_ref_data_mapping(self, legacy_object):
        obj_key = "_".join(legacy_object[k].strip() for k in self.mapped_legacy_keys)
        if obj_key in self.cache:
            return self.cache[obj_key]
        prepped_props = {k: legacy_object[k].strip() for k in self.mapped_legacy_keys}
        for mapping in self.regular_mappings:
            match_number = sum(prepped_props[k] == mapping[k] for k in self.mapped_legacy_keys)
            if match_number == len(self.mapped_legacy_keys):
                self.cache[obj_key] = mapping
                return mapping
        return None

    def is_hybrid_default_mapping(self, mapping):
        legacy_values = [value for key, value in mapping.items() if key in self.mapped_legacy_keys]
        return "*" in legacy_values and not self.is_default_mapping(mapping)

    def is_default_mapping(self, mapping):
        legacy_values = [value for key, value in mapping.items() if key in self.mapped_legacy_keys]
        return all(f == "*" for f in legacy_values)

    def pre_validate_map(self):
        if not any(f for f in self.map if f.get(f"folio_{self.key_type}", "")):
            raise TransformationProcessError(
                "", f"Column folio_{self.key_type} missing from {self.name} map file"
            )
        folio_values_from_map = [f[f"folio_{self.key_type}"] for f in self.map]
        folio_values_from_folio = [r[self.key_type] for r in self.ref_data]
        folio_values_not_in_map = list(
            {f for f in folio_values_from_folio if f not in folio_values_from_map}
        )
        map_values_not_in_folio = list(
            {f for f in folio_values_from_map if f not in folio_values_from_folio}
        )
        if any(folio_values_not_in_map):
            logging.info(
                "Values from %s ref data in FOLIO that are not in the map: %s",
                self.name,
                folio_values_not_in_map,
            )
        if any(map_values_not_in_folio):
            raise TransformationProcessError(
                "",
                (
                    f"Values ({self.key_type}) from {self.name} map are not in "
                    f"FOLIO: {map_values_not_in_folio}"
                ),
            )

    def post_validate_map(self):
        if not self.default_id:
            raise TransformationProcessError(
                "",
                f"No fallback {self.name} set up in map."
                f"Add a row to mapping file with *:s in all legacy "
                f"columns and a valid {self.name} value",
            )
        for mapping in self.map:
            if f"folio_{self.key_type}" not in mapping:
                logging.critical(
                    f"folio_{self.key_type} is not a column in the {self.name} mapping file. Fix."
                )
                sys.exit(1)
            elif (
                all(k not in mapping for k in self.mapped_legacy_keys)
                and "legacy_code" not in mapping
            ):
                logging.critical(
                    (
                        "field names from %s missing in map legacy_code is "
                        "not a column in the %s mapping file"
                    ),
                    self.mapped_legacy_keys,
                    self.name,
                )
                sys.exit(1)
            elif not all(mapping.values()):
                logging.critical(
                    f"empty value in mapping {mapping.values()}. Check {self.name} mapping file"
                )
                sys.exit(1)


def get_mapped_legacy_keys(mapping):
    return [
        k.strip()
        for k in mapping.keys()
        if k
        not in [
            "folio_group",
            "folio_code",
            "folio_id",
            "folio_name",
            "legacy_code",
            "folio_value",
            "folio_owner",
            "folio_feeFineType",
        ]
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['folio_migration_tools',
 'folio_migration_tools.mapping_file_transformation',
 'folio_migration_tools.marc_rules_transformation',
 'folio_migration_tools.migration_tasks',
 'folio_migration_tools.test_infrastructure',
 'folio_migration_tools.transaction_migration']

package_data = \
{'': ['*']}

install_requires = \
['argparse-prompt>=0.0.5,<0.0.6',
 'deepdiff>=6.2.3,<7.0.0',
 'defusedxml>=0.7.1,<0.8.0',
 'folio-uuid==0.2.9rc1',
 'folioclient>=0.50.1,<0.51.0',
 'httpx>=0.23.3,<0.24.0',
 'pyaml>=21.10.1,<22.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pyhumps>=3.7.3,<4.0.0',
 'pymarc>=5.0.0,<6.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-i18n>=0.3.9,<0.4.0']

setup_kwargs = {
    'name': 'folio-migration-tools',
    'version': '1.8.4rc3',
    'description': 'A tool allowing you to migrate data from legacy ILS:s (Library systems) into FOLIO LSP',
    'long_description': '# FOLIO Migration Tools\n![example workflow](https://github.com/FOLIO-FSE/MARC21-To-FOLIO/actions/workflows/python-app.yml/badge.svg)[![codecov](https://codecov.io/gh/FOLIO-FSE/folio_migration_tools/branch/main/graph/badge.svg?token=ZQL5ILWWGT)](https://codecov.io/gh/FOLIO-FSE/folio_migration_tools)   [![readthedocs](https://readthedocs.org/projects/docs/badge/?version=latest)](https://folio-migration-tools.readthedocs.io/)\n\nA toolkit that enables you to migrate data over from a legacy ILS system into [FOLIO LSP](https://www.folio.org/)\n\n# What is it good for?\nFOLIO Migration tools enables you to migrate libraries with the most common ILS:s over to FOLIO without data losses or any major data transformation tasks. \nThe tools transforms and loads the data providing you and the library with good actionable logs and data cleaning task lists together with the migrated data.\n\n## What data does it cover?\nFOLIO Migration Tools currently covers the following data sets:\n* Catalog (Inventory and SRS in FOLIO terminology)\n* Circulation transactions (Open loans and requests)\n* Users/Patrons (In FOLIO, these share the same app/database)\n* Courses and Reserves (Course reserves)\n* Organizations (Vendor records)\n* Orders (limited support)\n\n### What additional functionality is on the roadmap?\nThis is the loose roadmap, in order of most likely implementations first\n* ERM-related objects\n* Financial records\n\n### Can I use the tools for ongoing imports and integrations?\nThe tools are primarliy maintained for performing initial data migrations. We recommend that you use native FOLIO functionality for ongoing loads where possible. \nIn theory, these tools can be used for ongoing patron loads from systems like Banner, Workday, or PeopleSoft. But we recommend you to weigh your options carefully before going down this path. \n\n# Contributing\nWant to contribute? Read the [CONTRIBUTING.MD](https://github.com/FOLIO-FSE/folio_migration_tools/blob/main/CONTRIBUTING.md)\n\n# Found an issue?\nReport it on the [Github Issue tracker](https://github.com/FOLIO-FSE/folio_migration_tools/issues)\n\nThe scripts requires a FOLIO tenant with reference data properly set up. The script will throw messages telling what reference data is missing.\n# Installing\nMake sure you are running Python 3.9 or above. \n## 1. Using pip and venv\n### 2.1. Create and activate a [virtual environment](https://packaging.python.org/en/latest/guides/installing-using-pip-and-virtual-environments/#creating-a-virtual-environment)   \n```   \npython -m venv ./.venv     # Creates a virtual env in the current folder\nsource .venv/bin/activate  # Activates the venv    \n```\n### 2. Install using pip: \n```\npython -m pip install folio_migration_tools\n```\n### 3. Test the installation by showing the help pages \n```   \npython -m folio_migration_tools -h\n```    \n\n## 2. Using pipenv\n### 1. Run\n```   \npipenv install folio-migration-tools\n```   \n### 2. Test the installation by showing the help pages\n```  \npipenv run python3 -m folio_migration_tools -h\n```\n\n# FOLIO migration process\nThis repo plays the main part in a process using a collection of tools. The process itself is documented in more detail, including example configuration files, at [this template repository](https://github.com/FOLIO-FSE/migration_repo_template)\nIn order to perform migrations according to this process, you need the following:\n* An Installation of [FOLIO Migration Tools](https://pypi.org/project/folio-migration-tools/). Installation instructions above.\n* A clone, or a separate repo created from [migration_repo_template](https://github.com/FOLIO-FSE/migration_repo_template)\n* Access to the [Data mapping file creator](https://data-mapping-file-creator.folio.ebsco.com/data_mapping_creation) web tool\n* A FOLIO tenant running the latest or the second latest version of FOLIO\n\n# Internationalization\n\nThis repo uses [Python-i18n](https://github.com/danhper/python-i18n) to translate reports between languages, and to handle large strings for templates.\n\n**Any English string which will end up in a report** should be wrapped in the function `i18n.t` from `i18n`:\n\n## Keys/Usage\n\n```js\ni18n.t("Reports")+":"\n```\n\nTemplating is achieved with `%{[key]}` blocks, and keyword arguments in the internationaliation:\n\n```js\ni18n.t("Code \'%{code}\' not found in FOLIO",code=folio_code)\n```\n\nLong strings can use a placeholder key:\n\n```js\ni18n.t("blurbs.Introduction.description")\n```\n\nWith the full string in ```translations/en.json```:\n\n```json\n"blurbs.Introduction.description": "<br/>Data errors preventing records from being migrated\n```\n\n## Translations Files\n\nTranslation files live in the `translations` directory, with `en.json` as the default.\n\nExtract template files with the `extract_translations` script:\n\n```bash\npython scripts/extract_translations.py\n```\n\n## Internationalizations\n\nOther langauges translations live in `translations/[locale].json`.\nFor example, Spanish would be `es.json`. \n\nThe keys must match the English keys, but the Values should be translated.\n\nYou can update a language file\'s keys with:\n\n```bash\npython scripts/update_language.py --target-lang [locale]\n```\n\nTranslate all new strings, which begin with `TRANSLATE`, then commit.\n\n## Tips\n\n* Internationalize entire phrases or paragraphs, not just the constitutent words. Syntax and grammar vary significantly between languages.\n* Name template variables as generically as possible in the circumstance, and check translations for reusable translations.\n* In a block with sentences separately followed by values, such as a table, you only need to translate the sentences. \n\n# Running the scripts\nFor information on syntax, what files are needed and produced by the toolkit, refer to the documentation and example files in the [template repository](https://github.com/FOLIO-FSE/migration_repo_template). We are building out the docs section in this repository as well:[Documentation](https://folio-migration-tools.readthedocs.io/en/latest/)\n¨\n',
    'author': 'Theodor Tolstoy',
    'author_email': 'github.teddes@tolstoy.se',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/FOLIO-FSE/folio_migration_tools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
import i18n


class TransfomationError(Exception):
    pass


class TransformationFieldMappingError(TransfomationError):
    """Raised when the a field mapping fails, but the error is not critical.
    The issue should be logged for the library to act upon it"""

    def __init__(self, index_or_id="", message="", data_value=""):
        self.index_or_id = index_or_id or ""
        self.message = message
        self.data_value = data_value
        super().__init__(self.message)

    def __str__(self):
        return (
            i18n.t("Data issue. Consider fixing the record. ")
            + f"\t{self.index_or_id}\t{self.message}\t{self.data_value}"
        )

    def log_it(self):
        logging.log(
            26,
            "FIELD MAPPING FAILED\t%s\t%s\t%s",
            self.index_or_id,
            self.message,
            self.data_value,
        )


class TransformationRecordFailedError(TransfomationError):
    """Raised when the a field mapping fails, Error is critical and means tranformation fails"""

    def __init__(self, index_or_id, message="", data_value=""):
        self.index_or_id = index_or_id
        self.message = message
        self.data_value = data_value
        # logging.log(26, f"RECORD FAILED\t{self.id}\t{self.message}\t{self.data_value}")
        super().__init__(self.message)

    def __str__(self):
        return (
            f"Critical data issue. Record needs fixing"
            f"\t{self.index_or_id}\t{self.message}\t{self.data_value}"
        )

    def log_it(self):
        logging.log(
            26,
            "RECORD FAILED\t%s\t%s\t%s",
            self.index_or_id,
            self.message,
            self.data_value,
        )


class TransformationProcessError(TransfomationError):
    """Raised when the transformation fails due to incorrect configuraiton,
    mapping or reference data. This error should take the process to a halt."""

    def __init__(
        self,
        index_or_id,
        message="Critical Process issue. Transformation failed."
        " Check configuration, mapping files and reference data",
        data_value="",
    ):
        self.index_or_id = index_or_id
        self.message = message
        self.data_value = data_value
        super().__init__(self.message)

    def __str__(self):
        return (
            f"Critical Process issue. Check configuration, mapping files and reference data"
            f"\t{self.index_or_id}\t{self.message}\t{self.data_value}"
        )
